from __future__ import absolute_import, division, print_function, unicode_literals

from math import pi, cos, sin
from itertools import islice

import numpy as np
import cv2
from sklearn.cluster import MeanShift

from .rand import rgb_colors
from .math import degree_to_rad, inf

## UNFINISHED
## BASED ON: Resizing by Symmetry-Summarization (2010)

def MSER_draw(img, boxes, centers, major_v, minor_v, labels):

	center_int = centers.astype(int)
	major_int = (centers + major_v).astype(int)
	minor_int = (centers + minor_v).astype(int)

	labels_unique = np.unique(labels)
	n_clusters = len(labels_unique)
	colors = list(islice(rgb_colors(), n_clusters))
	print("{} clusters found".format(n_clusters))

	for box, label in zip(boxes, labels):
		cv2.ellipse(img, box, colors[label], 2)

	for c, ma, mi in zip(center_int, major_int, minor_int):
		cv2.line(img, tuple(c), tuple(ma), (0, 255, 0), 2)
		cv2.line(img, tuple(c), tuple(mi), (0, 255, 0), 2)

def MSER_boxes(img):
	mser = cv2.MSER_create()

	msers, _ = mser.detectRegions(img)

	return [cv2.fitEllipse(mser) for mser in msers]

def MSER_ellipses(boxes):
	pihalf = pi/2.
	boxes = np.array([(x, y, ma, mi, a) for (x, y), (ma, mi), a in boxes])

	centers = boxes[:,0:2]
	angles = boxes[:,4]*2.*pi/360. # convert degrees to rad

	major_x = boxes[:,2]/2 * np.cos(angles)
	major_y = boxes[:,2]/2 * np.sin(angles)
	major_v = np.stack([major_x, major_y], axis=-1)

	angles += pihalf
	minor_x = boxes[:,3]/2 * np.cos(angles)
	minor_y = boxes[:,3]/2 * np.sin(angles)
	minor_v = np.stack([minor_x, minor_y], axis=-1)

	return (centers, major_v, minor_v)

def MSER(img, show=True):
	boxes = MSER_boxes(img)
	centers, major_v, minor_v = MSER_ellipses(boxes)

	return boxes, centers, major_v, minor_v

def norm(v):
	return np.sqrt(np.sum(v**2, axis=-1))

def shape_dissimmilarity(R_u, R_v, i, j):
	num1 = np.abs(R_u[i] - R_u[j])
	denom1 = max(norm(R_u[i]), norm(R_u[j]))
	
	num2 = np.abs(R_v[i] - R_v[j])
	denom2 = max(norm(R_v[i]), norm(R_v[j]))

	return num1/denom1 + num2/denom2

def appearance_dissimmilarity():
	pass

def symmetry_detection(image, K_c=5):

	#Identifying cells with MSER
	boxes, R_c, R_u, R_v = MSER(image) # {r_i} set of MSERs (2d ellipsis) r_i = (center, major axis, minor axis)

	# Clustering with adaptive mean-shift clustering
	u_norm = norm(R_u)
	v_norm = norm(R_v)
	shapessizes = np.stack([u_norm, v_norm], axis=-1)

	ms = MeanShift()
	labels = ms.fit_predict(shapessizes)

	MSER_draw(image, boxes, R_c, R_u, R_v, labels)
	cv2.imshow('img', image)
	cv2.waitKey()

def diff(v1, v2, tau_v):
	val = abs(v1-v2)/max(norm(v1), norm(v2))
	if val > tau_v:
		return inf
	else:
		return val

if __name__ == "__main__":
	imgpath = "D:\\colloseo.jpg"
	img = cv2.imread(imgpath)
	symmetry_detection(img)



from __future__ import absolute_import, division, print_function, unicode_literals

import platform
from collections import namedtuple

_usagetuple = namedtuple("usage", "total used free")
_volumeinfotuple = namedtuple("volumeinfo", "VolumeName VolumeSerialNumber MaximumComponentLength FileSystemFlags FileSystemName")

def is_os_64bit():
	return platform.machine().endswith("64")



from __future__ import absolute_import, division, print_function, unicode_literals

from math import log2, exp
from itertools import chain
from collections import Counter
from sys import stderr
from typing import TYPE_CHECKING

import numpy as np

from .numpy import normalize, categorical, batchtopk
from .datastructures import VariableRowMatrix
from .encoder import BatchLabelEncoder
from .iter import count_distinct, progress
from .sequence import LasyStringList
from .object import cast

if TYPE_CHECKING:
	from typing import Any, Dict, Iterable
	Documents = List[List[int]]
	IterableDocuments = Iterable[Iterable[int]]

# utils

def mylen(obj):
	from scipy.sparse import issparse

	if issparse(obj):
		return obj.shape[0]
	else:
		return len(obj)

def top_topics(id2word, term_topic_matrix, num_words=10, decimals=2):
	# type: (Dict[int, str], float[K, V], int) -> Iterator[List[Tuple[str, float]]]

	indices_list, probs_list = batchtopk(term_topic_matrix, num_words, reverse=True)
	np.around(probs_list, decimals=2, out=probs_list)

	for indices, probs in zip(indices_list, probs_list):
		yield [(id2word[id], prob) for id, prob in zip(indices, probs)]

def format_topics(topics, linesep="\n", tokensep="\t"):
	# type: (iterable[List[Tuple[str, float]]], ) -> str

	buffer = []

	for terms in topics:
		buffer.append(tokensep.join("({},{})".format(term, prob) for term, prob in terms))

	return linesep.join(buffer)

class LDADocument(object):

	def __init__(self, words):
		# type: (List[int], ) -> None

		self.words = words

	def __iter__(self):
		# type: () -> Iterator[int]

		return iter(self.words)

	def __len__(self):
		# type: () -> int

		return len(self.words)

	def __repr__(self):
		# type: () -> str

		return "LDADocument({})".format(repr(self.words))

class LDABase(object):

	def __init__(self):
		# type: () -> None

		np.seterr(all="raise")

	def init_topic(self):
		# type: () -> int

		return np.random.randint(self.K)

	def add_counts(self, m, t, k):
		# type: (int, int, int) -> None

		self.nmk[m, k] += 1
		self.nkt[k, t] += 1
		self.nk[k] += 1

	def sub_counts(self, m, t, k):
		# type: (int, int, int) -> None

		self.nmk[m, k] -= 1
		self.nkt[k, t] -= 1
		self.nk[k] -= 1

	def _validate(self, M, K, V):
		assert M > 0, "Number of documents must be larger than zero"
		assert K > 0, "Number of topics must be larger than zero"
		assert V > 0, "Vocabulary size must be larger than zero"

		assert self.nmk.shape == (M, K)
		assert self.nm.shape == (M, )
		assert self.nkt.shape == (K, V)
		assert self.nk.shape == (K, )

		assert self.α.shape == (K, )
		assert self.β.shape == (V, )

	def _initialize_topics(self, docs, topics):
		# type: (IterableDocuments, Dict[Any, int]) -> None

		for m, doc in enumerate(progress(docs, file=stderr)):
			for i, t in enumerate(doc):
				k = topics[m, i] = self.init_topic()
				self.add_counts(m, t, k)

	def calc_probs(self, m, t):
		# type: (int, int) -> float[K]

		# notes: np.sum(self.nkt, axis=-1) == self.nk
		# notes: np.sum(self.nmk[m], axis=-1) == self.nm[m]

		""" This is eq1. in the DC-LDA paper. """

		""" The α and β terms could be moved to `initialize`.
		"""

		""" The full term for left would be:
			left = (self.nmk[m, :] + self.α) / (self.nm[m] + self.αsum) # [K]
			however, the denominator can be dropped, because it doesn't depend on `k`
			and we only care about proportionality.
			It is done so in: 'A Theoretical and Practical Implementation Tutorial
				on Topic Modeling and Gibbs Sampling'
		"""

		left = (self.nmk[m, :] + self.α) # [K]
		right = (self.nkt[:, t] + self.β[t]) / (self.nk[:] + self.βsum) # [K]
		return left * right # [K]

	def calc_probs_selected(self, m, t, k_select):
		# type: (int, int) -> float[K]

		left = (self.nmk[m, k_select] + self.α) # [K]
		right = (self.nkt[k_select, t] + self.β[t]) / (self.nk[k_select] + self.βsum) # [K]
		return left * right # [K]

	def sample(self, m, t):
		# type: (int, int) -> int

		probs = self.calc_probs(m, t) # [K]
		return categorical(normalize(probs))

	def sample_selected(self, m, t, k_select):
		# type: (int, int, Any) -> int

		raise NotImplementedError

	def _sample_all(self, docs, topics):
		# type: (IterableDocuments, Dict[Any, int]) -> None

		for m, doc in enumerate(progress(docs, file=stderr)):
			for i, t in enumerate(doc):
				k = topics[m, i]
				self.sub_counts(m, t, k)
				k = topics[m, i] = self.sample(m, t, k) # sample new topics
				self.add_counts(m, t, k)

	def initialize(self):
		# type: () -> None

		""" Set parameters based on input docs """

		raise NotImplementedError

	def initialize_topics(self): # step1
		# type: () -> None

		""" Initialize topics matrix """

		raise NotImplementedError

	def sample_all(self): # step2 and step3
		# type: () -> None

		""" Sample new topics """

		raise NotImplementedError

	def metric(self):
		# type: () -> float

		raise NotImplementedError

	def fit(self, n_iter=100, verbose=False):
		# type: (int, bool) -> None

		""" Fit LDA """

		assert self.docs, "No documents added to model"

		self.initialize()
		self.initialize_docs()
		self.initialize_topics()

		for i in range(n_iter):
			self.sample_all()
			if verbose:
				print("Step #{}, metric: {}".format(i, self.metric()))

	def theta(self):
		# type: () -> float[M, K]

		num = self.nmk + self.α # [M, K]
		denom = self.nm + self.αsum # [M]
		return num / denom[:,None] # [M, K]

	def phi(self):
		# type: () -> float[K, V]

		num = self.nkt + self.β # [K, V]
		denom = self.nk + self.βsum # [K]
		return num / denom[:,None] # [K, V]

	def _perplexity(self, docs):
		# type: (IterableDocuments, ) -> float

		θ = self.theta()
		φ = self.phi()

		num = 0
		denom = 0

		for m, doc in enumerate(docs):
			for i, t in enumerate(doc):
				num += np.log(np.inner(φ[:,t], θ[m]))
			denom += len(doc)

		return exp(-num / denom)

	def docs2topics(self):
		# type: () -> int[M]

		return np.argmax(self.theta(), axis=-1) # [M]

	def print_topics(self, num_words=10):
		# type: (int, ) -> List[list[Tuple[str, float]]]

		topics = top_topics(self.id2word, self.phi(), num_words)
		print(format_topics(topics))

	@staticmethod
	def generate(theta, phi, doc_lens):
		# type: (float[K, V], float[M, K], Iterable[int]) -> Tuple[dict, dict]

		z = {} # sparse
		w = {} # sparse

		for m, length in enumerate(doc_lens):
			for i in range(length):
				k = z[(m, i)] = categorical(theta[m]) # scalar # doc to topic
				w[(m, i)] = categorical(phi[k]) # scalar # topic to word

		return z, w

	def inferencer(self):
		# type: () -> LDAInfer

		return cast(self, LDAInfer)

class LDA(LDABase):

	""" LDA using collapsed Gibbs sampling
		Based on:
		- https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation
		- Latent Dirichlet Allocation (2003)
		- A Theoretical and Practical Implementation Tutorial on Topic Modeling and Gibbs Sampling (2011)
		- Integrating Out Multinomial Parameters in Latent Dirichlet Allocation and Naive Bayes for Collapsed Gibbs Sampling (2010)
	"""

	def __init__(self, n_topics, alpha=1., beta=1.):
		# type: (int, int, float, float) -> None

		LDABase.__init__(self)

		self.K = n_topics # number of topics

		self.alpha = alpha
		self.beta = beta

		self.word_encoder = BatchLabelEncoder(tokenizer="nltk")
		self.docs = [] # type: List[LDADocument]

		self.inttype = np.int32

	def initialize(self):
		# type: () -> None

		self.V = self.word_encoder.num_labels

		self.α = np.full(self.K, self.alpha / self.K) # [K] symmetric Dirichlet prior for document-topic distribution θ
		self.β = np.full(self.V, self.beta / self.V) # [V] symmetric Dirichlet prior for topic-word distribution φ

		self.αsum = np.sum(self.α)
		self.βsum = np.sum(self.β)

		self.nkt = np.zeros((self.K, self.V), dtype=self.inttype) # [K, V] topic-word counts
		self.nk = np.zeros((self.K, ), dtype=self.inttype) # [K] total topic counts

	def initialize_docs(self):
		# type: () -> None

		self.M = len(self.docs)

		self.nmk = np.zeros((self.M, self.K), dtype=self.inttype) # [M, K] document-topic counts
		self.nm = np.array(list(map(len, self.docs)), dtype=self.inttype) # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

		self.topics = {} # `z`, sparse document-word topics matrix (because rows can have different lengths)

	def initialize_topics(self):
		# type: () -> None

		self._validate(self.M, self.K, self.V)

		return self._initialize_topics(self.docs, self.topics)

	def sample_all(self):
		# type: () -> None

		return self._sample_all(self.docs, self.topics)

	def metric(self):
		# type: () -> float

		return self._perplexity(self.docs)

	def perplexity(self, docs):
		# type: (Iterable[List[int]], ) -> float

		return self._perplexity(docs)

	def transform(self, doc_lens):
		# type: (Iterable[int], ) -> Tuple[dict, dict]

		return self.generate(self.theta(), self.phi(), doc_lens)

	def add_doc(self, doc):
		# type: (str, ) -> int

		assert isinstance(doc, str)

		words = list(self.word_encoder.fit_transform(doc))

		ret = len(self.docs)
		self.docs.append(LDADocument(words))
		return ret

	def make_doc(self, doc):
		# type: (str, ) -> LDADocument

		assert isinstance(doc, str)

		words = list(self.word_encoder.transform(doc))
		return LDADocument(words)

class LDATermWeight(LDABase):

	""" LDA with term weighting
		Based on: "Term Weighting Schemes for Latent Dirichlet Allocation" (2010)

		Further analysis papers:
			- "Pulling Out the Stops: Rethinking Stopword Removal for Topic Models" (2017)
			- "Assessing topic model relevance: Evaluation and informative priors" (2019)
	"""

	def __init__(self, n_topics, tws="PMI", alpha=1., beta=1.):
		# type: (int, int, str, float, float) -> None

		""" tws: Term weighting scheme
			ONE: Consider every term equal (same as standard LDA)
			TF: Term frequency
			IDF: Inverse Document Frequency term weighting.
				Thus, a term occurring at almost every document has very low weighting and a term occurring at a few document has high weighting.
			PMI: Pointwise Mutual Information term weighting (default)
		"""

		LDABase.__init__(self)

		self.K = n_topics # number of topics

		self.alpha = alpha
		self.beta = beta

		self.word_encoder = BatchLabelEncoder(tokenizer="nltk")
		self.docs = [] # type: List[LDADocument]

		self.create_term_weights = {
			"ONE": self._one,
			"TF": self._tf,
			"IDF": self._idf,
			"TFIDF": self._tfidf,
			"PMI": self._pmi,
			"RAND": self._rand,
		}[tws]

		self.floattype = np.float32
		self.inttype = np.int32

	def _one(self, docs):
		# type: (Any, ) -> int[M, V]

		# return np.ones((self.M, self.V))
		return np.broadcast_to(1, (self.M, self.V)) # is this slower for indexing later?

	def _tf(self, docs):
		# type: (IterableDocuments, ) -> float[M, V]

		all_counts = Counter(chain.from_iterable(docs))
		num_counts = sum(all_counts.values())

		tw = np.array([all_counts[t] for t in range(self.V)], dtype=self.floattype) # [V]
		tw = -np.log2(tw / num_counts)

		return np.broadcast_to(tw, (self.M, self.V))

	def _idf(self, docs):
		# type: (IterableDocuments, ) -> float[M, V]

		tw = np.zeros((self.V, ), dtype=self.floattype) # [V]

		for doc in docs:
			for token in set(doc):
				tw[token] += 1

		tw = np.log2(self.M / tw)

		return np.broadcast_to(tw, (self.M, self.V))

	def _tfidf(self, docs):
		raise NotImplementedError

	def _pmi(self, docs):
		# type: (IterableDocuments, ) -> float[M, V]

		tw = np.zeros((self.M, self.V), dtype=self.floattype) # [M, V] term weights

		all_counts = Counter(chain.from_iterable(docs))

		for m, doc in enumerate(docs):
			doc_counts = Counter(doc)
			for i, t in enumerate(doc):
				tw[m, t] = -log2(doc_counts[t] / all_counts[t])

		np.clip(tw, 0, None, out=tw)
		return tw

	def _rand(self, docs):
		# type: (Any, ) -> float[M, V]

		return np.random.uniform(0, 1, (self.M, self.V)).astype(np.float32)

	def initialize(self):
		# type: () -> None

		self.V = self.word_encoder.num_labels

		self.α = np.full(self.K, self.alpha / self.K) # [K] symmetric Dirichlet prior for document-topic distribution θ
		self.β = np.full(self.V, self.beta / self.V) # [V] symmetric Dirichlet prior for topic-word distribution φ

		self.αsum = np.sum(self.α)
		self.βsum = np.sum(self.β)

		self.nkt = np.zeros((self.K, self.V), dtype=self.inttype) # [K, V] counts, same as np.sum(nmkt, axis=0)
		self.nk = np.zeros((self.K, ), dtype=self.inttype) # [K] counts, same as np.sum(nkt, axis=-1) # [K]

		#self.wkt = np.zeros((self.K, self.V), dtype=self.floattype) # [K, T] weights
		#self.wk = np.zeros((self.K, ), dtype=self.floattype) # [K] weights

	def initialize_docs(self):
		# type: () -> None

		self.M = len(self.docs)

		self.tw = self.create_term_weights(self.docs)
		print(self.tw.shape)
		print(self.tw)

		self.nmk = np.zeros((self.M, self.K), dtype=self.inttype) # [M, K] counts, same as np.sum(nmkt, axis=-1)
		self.nm = np.array(list(map(len, self.docs)), dtype=self.inttype) # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

		"""
			`wmk[m]` is same as `np.sum(wmkt[m], axis=-1)` is same as
			`np.sum(self.tw[m, :][None, :] * nmkt[m], axis=-1)`

			with
				`wmkt = np.zeros((self.M, self.K, self.V), dtype=self.floattype)` # [M, K, V] weights
				`nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)` # [M, K, V] counts
		"""
		#self.nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)
		self.wmk = np.zeros((self.M, self.K), dtype=self.floattype) # [M, K] weights, same as np.sum(nmkt, axis=-1)

		self.topics = VariableRowMatrix() # [M, Variable] `z`, sparse document-word topics matrix

	def initialize_topics(self):
		# type: () -> None

		# self._validate(self.M, self.K, self.V)

		self._initialize_topics(self.docs, self.topics)

	def calc_probs(self, m, t):
		# type: (int, int) -> float[K]

		# here using global weights works
		#wmk = np.sum(self.tw[m, :][None, :] * self.nmkt[m], axis=-1, dtype=self.floattype)
		#if not np.allclose(self.wmk[m], wmk, atol=1e-06): # after some time it still diverges because of nummerical issues
		#	raise RuntimeError("inconsistent wmk")

		# here it doesn't...
		wkt = self.tw[m, t] * self.nkt[:, t]
		#if not np.allclose(self.wkt[:, t], wkt[:]):
		#	raise RuntimeError("inconsistent wkt")

		# here it doesn't also...
		wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1) # not the same as `self.wk[k] (+|-)= self.tw[m, t]`
		#if not np.allclose(self.wk, wk):
		#	raise RuntimeError("inconsistent wk")

		# nk: ld.numByTopic[tid]
		# nkt: ld.numByTopicWord[tid, vid]

		left = (self.wmk[m] + self.α) # [K]
		right = (wkt[:] + self.β[t]) / (wk[:] + self.βsum) # [K]
		return left * right # [K]

	"""
	def addWordToINC(docId, tid, vid):

		# paper W*beta term in eq. 6 (https://www.aclweb.org/anthology/N10-1070.pdf)
		ld.numByTopicWordCOUNTS[tid, vid] += 1
		Nk = np.sum(docs[docId].wordWeights[None,:] * ld.numByTopicWordCOUNTS, axis=-1) # [None,:] needed for correct broadcasting. summed over words, K size array remains
		# or the above in plain python
		Nk = [sum(docs[docId].wordWeights[word_id] * ld.numByTopicWordCOUNTS[tid, word_id]
			for word_id in docs[docid].words)
				for tid in range(K)] # K size array

		# addWordTo INC
		ld.numByTopic[tid] += docs[docId].wordWeights[pid] # ld.numByTopic is a K size array

		# now both weight sums should be the same AT ALL TIMES
		assert Nk == ld.numByTopic # elementwise comparison

		# but they are not after the first document, if the weights depend on the document
	"""

	def add_counts(self, m, t, k):
		# type: (int, int, int) -> None

		self.nmk[m, k] += 1
		self.nkt[k, t] += 1
		self.nk[k] += 1

		#self.nmkt[m, k, t] += 1
		self.wmk[m, k] += self.tw[m, t]
		#self.wkt[k, t] += self.tw[m, t]
		#self.wk[k] += self.tw[m, t]

		"""
		print("docId={}, vid={}, tid={}".format(m, t, k))

		wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1)
		print("ld.numByTopicWordCOUNTS:\n", self.nkt)
		print("paper sum:", wk)
		print("ld.numByTopic:", self.wk)

		if not np.allclose(wk, self.wk):
			raise RuntimeError("not equal")
		"""

	def sub_counts(self, m, t, k):
		# type: (int, int, int) -> None

		self.nmk[m, k] -= 1
		self.nkt[k, t] -= 1
		self.nk[k] -= 1

		#self.nmkt[m, k, t] -= 1
		self.wmk[m, k] -= self.tw[m, t]
		#self.wkt[k, t] -= self.tw[m, t]
		#self.wk[k] -= self.tw[m, t]

	def sample_all(self):
		# type: () -> None

		self._sample_all(self.docs, self.topics)

	def metric(self):
		# type: () -> float

		return self._perplexity(self.docs)

class LDAInfer(LDA):

	def __init__(self):
		# type: () -> None

		self.nkt_old = self.nkt
		self.nk_old = self.nk
		self.docs = []

	def calc_probs(self, m, t):
		# type: (int, int) -> float[]

		left = (self.nmk[m, :] + self.α) # [K]
		right = (self.nkt[:, t] + self.nkt_old[:, t] + self.β) / (self.nk[:] + self.nk_old[:] + self.βsum) # [K]

		return left * right # [K]

	def initialize(self):
		# type: () -> None

		self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)
		self.nk = np.zeros((self.K, ), dtype=self.inttype)

if __name__ == "__main__":
	from sklearn.datasets import fetch_20newsgroups
	from .pickle import cache

	ng = fetch_20newsgroups()

	@cache("lda.init.p.gz")
	def init_lda():
		lda = LDA(n_topics=20)

		for doc in ng.data:
			lda.add_doc(doc)

		return lda

	@cache("lda_tw.init.p.gz")
	def init_lda_tw():
		lda = LDATermWeight(n_topics=20)

		for doc in ng.data:
			lda.add_doc(doc)

		return lda

	debug_docs = [
		[0, 6, 3, 9, 1, 5],
		[9, 8, 8, 7, 1, 0],
		[2, 4, 4, 5, 0, 1],
	]

	print("LDA")
	lda = LDATermWeight(2, tws="RAND")
	lda.add_docs_encoded()
	lda.fit(10, docs=debug_docs, verbose=True)
	lda.print_topics()

	print("LDA")
	lda = init_lda()
	lda.fit(10, verbose=True)

	print("LDATermWeight")
	lda_tw = init_lda_tw()
	lda.fit(2, verbose=True)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Any, Callable, Sequence, TypeVar
	T = TypeVar("T")

def extract_indexes(d):
	# type: (dict, ) -> Callable([Sequence], dict)

	return lambda seq: {name: seq[index] for index, name in viewitems(d)}

def bit_or(x, y):
	# type: (Any, Any) -> Any

	return x|y

def not_in(s):
	# type: (Container[T], ) -> Callable[[T], bool]

	return lambda x: x not in s

def logical_xor(a, b):
	# type: (Any, Any) -> bool
	# this is just the != operator for bools

	return not (a and b) and (a or b)

def logical_implication(a, b):
	# type: (Any, Any) -> bool

	""" if a is True, b must be True """

	return (not a) or b

def converse_implication(a, b):
	# type: (Any, Any) -> bool

	""" if a is False, b must be False """

	return a or (not b)

def converse_nonimplication(a, b):
	# type: (Any, Any) -> bool

	""" if a is False, b must be True """

	return (not a) and b



from __future__ import absolute_import, division, print_function, unicode_literals

from time import time
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Any, Callable, Hashable, Iterator, Optional, Tuple, TypeVar
	T = TypeVar("T")

def iter_timer(it):
	# type: (Iterator[T], ) -> Iterator[Tuple[T, float]]
	try:
		while True:
			start = time()
			res = next(it)
			delta = time()-start
			yield res, delta
	except StopIteration:
		pass

class DeltaTime(object):

	__slots__ = ("start", "end")

	def __init__(self):
		# type: () -> None

		self.start = None # type: Optional[float]
		self.end = time()

	def __iter__(self):
		# type: () -> DeltaTime

		return self

	def __next__(self):
		# type: () -> float

		self.start, self.end = self.end, time()
		return self.end - self.start

	def get(self):
		# type: () -> float

		return time() - self.end

	def reset(self):
		# type: () -> None

		self.end = time()

	def get_reset(self):
		# type: () -> float

		self.start, self.end = self.end, time()
		return self.end - self.start

class PrintStatementTime(object):

	__slots__ = ("tpl", "start")

	def __init__(self, tpl=None):
		# type: (Optional[str], ) -> None
		if tpl is None:
			self.tpl = "Execution took {delta}s"
		else:
			self.tpl = tpl

	def __enter__(self):
		self.start = time()
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		end = time()
		delta = end - self.start
		msg = self.tpl.format(start=self.start, end=end, delta=delta)
		if exc_type:
			print(msg + " (interrupted)")
		else:
			print(msg)

class MeasureTime(object):

	__slots__ = ("delta", "start")

	def __init__(self):
		# type: () -> None
		self.delta = None

	def __enter__(self):
		self.start = time()
		return self

	def __exit__(self, type, value, traceback):
		self.delta = time() - self.start

	def get(self):
		# type: () -> float
		if self.delta:
			return self.delta
		else:
			return time() - self.start

class TimeIt(object):

	def __init__(self):
		self.results = defaultdict(list)
		self.starts = dict()

	def __call__(self, key, func, *args, **kwargs):
		# type: (Hashable, Callable, *Any, **Any) -> Any

		with MeasureTime() as t:
			ret = func(*args, **kwargs)
		self.results[key].append(t.get())
		return ret

	def start(self, key):
		# type: (Hashable, ) -> None
		self.starts[key] = DeltaTime()

	def stop(self, key):
		# type: (Hashable, ) -> None
		self.results[key].append(self.starts[key].get())

	def min(self, key):
		# type: (Hashable, ) -> float
		return min(self.results[key])

	def length(self, key):
		# type: (Hashable, ) -> int
		return len(self.results[key])



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys, viewvalues
from itertools import chain, repeat

def upsert(cursor, primary, values, table):
	# type: (Cursor, dict, dict, str) -> bool

	# use INSERT ... ON DUPLICATE KEY UPDATE instead?

	set_str = ",".join("{}=?".format(k) for k in viewkeys(values))
	where_str = " AND ".join("{}=?".format(k) for k in viewkeys(primary))
	cursor.execute("UPDATE {} SET {} WHERE {}".format(table, set_str, where_str), chain(viewvalues(values), viewvalues(primary)))

	if cursor.rowcount == 0:
		into_str = ",".join(chain(viewkeys(primary), viewkeys(values)))
		values_str = ",".join(repeat("?", len(primary) + len(values)))
		cursor.execute("INSERT INTO {} ({}) VALUES ({})".format(table, into_str, values_str), chain(viewvalues(primary), viewvalues(values)))
		return True
	return False



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range

class RingList:

	""" bad complexity, moves all data if ring is full """

	def __init__(self, length):
		self.data = []
		self.full = False
		self.max = length
		self.cur = 0

	def append(self, x):
		if self.full:
			for i in range(self.cur-1):
				self.data[i] = self.data[i+1]
			self.data[self.cur-1] = x
		else:
			self.data.append(x)
			self.cur += 1
			if self.cur == self.max:
				self.full = True

	def get(self):
		return self.data

	def remove(self):
		if self.cur > 0:
			del self.data[self.cur-1]
			self.cur -= 1

	def size(self):
		return self.cur

	def maxsize(self):
		return self.max

	def __str__(self):
		return "".join(self.data)



from __future__ import absolute_import, division, print_function, unicode_literals

from sys import stdout
from functools import wraps, reduce
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Callable, Any, TextIO

def identity(x):
	# type: (T, ) -> T
	""" Identity function. """

	return x

def nop():
	# type: () -> None
	""" Function which does absolutely nothing (aka pass, noop). """

	pass

def compose_two(f, g):
	""" compose_two(f, g) -> lambda x: f(g(x)) """

	return lambda x: f(g(x))

def compose(*functions):
	# type: (Iterable[Callable[[Any], Any]], ) -> Callable[[Any], Any]

	""" compose(f, g, h) -> lambda x: f(g(h(x))) """

	return reduce(compose_two, functions, identity)

def apply(f, x): # *args, **kwargs
	# type: (Callable[[T], U], T) -> U

	return f(x)

def zipmap(funcs, vals):
	# type: (Iterable[Callable[[T], U]], Iterable[T]) -> Iterator[U]

	""" applies a list of functions to a list of values """

	return map(apply, funcs, vals)

def multiapply(funcs, elm):
	for func in funcs:
		elm = func(elm)
	return elm

def multimap(funcs, it):
	for i in it:
		yield multiapply(funcs, i)

def call_repeated(num):
	# type: (int, ) -> Callable[[Callable], Callable]

	""" Function decorator to call decorated function `num` times with the same arguments.
		Returns the results of the last call.
	"""

	assert num > 0

	def dec(func):
		@wraps(func)
		def inner(*args, **kwargs):
			last = None
			for i in range(num):
				last = func(*args, **kwargs)
			return last
		return inner
	return dec

def print_return_type(func, file=stdout):
	# type: (Callable, TextIO) -> Callable

	""" Wraps function to print the return type after calling. """

	def inner(*args, **kwargs):
		ret = func(*args, **kwargs)
		print(type(ret), file=file)
		return ret
	return inner



from __future__ import absolute_import, division, print_function, unicode_literals

FILE_DEVICE_SCSI = 0x0000001b

IOCTL_SCSI_EXECUTE_IN = (FILE_DEVICE_SCSI << 16) + 0x0011
IOCTL_SCSI_EXECUTE_OUT = (FILE_DEVICE_SCSI << 16) + 0x0012
IOCTL_SCSI_EXECUTE_NONE = (FILE_DEVICE_SCSI << 16) + 0x0013

# SMART support in atapi

IOCTL_SCSI_MINIPORT_SMART_VERSION = (FILE_DEVICE_SCSI << 16) + 0x0500
IOCTL_SCSI_MINIPORT_IDENTIFY = (FILE_DEVICE_SCSI << 16) + 0x0501
IOCTL_SCSI_MINIPORT_READ_SMART_ATTRIBS = (FILE_DEVICE_SCSI << 16) + 0x0502
IOCTL_SCSI_MINIPORT_READ_SMART_THRESHOLDS = (FILE_DEVICE_SCSI << 16) + 0x0503
IOCTL_SCSI_MINIPORT_ENABLE_SMART = (FILE_DEVICE_SCSI << 16) + 0x0504
IOCTL_SCSI_MINIPORT_DISABLE_SMART = (FILE_DEVICE_SCSI << 16) + 0x0505
IOCTL_SCSI_MINIPORT_RETURN_STATUS = (FILE_DEVICE_SCSI << 16) + 0x0506
IOCTL_SCSI_MINIPORT_ENABLE_DISABLE_AUTOSAVE = (FILE_DEVICE_SCSI << 16) + 0x0507
IOCTL_SCSI_MINIPORT_SAVE_ATTRIBUTE_VALUES = (FILE_DEVICE_SCSI << 16) + 0x0508
IOCTL_SCSI_MINIPORT_EXECUTE_OFFLINE_DIAGS = (FILE_DEVICE_SCSI << 16) + 0x0509
IOCTL_SCSI_MINIPORT_ENABLE_DISABLE_AUTO_OFFLINE = (FILE_DEVICE_SCSI << 16) + 0x050a
IOCTL_SCSI_MINIPORT_READ_SMART_LOG = (FILE_DEVICE_SCSI << 16) + 0x050b
IOCTL_SCSI_MINIPORT_WRITE_SMART_LOG = (FILE_DEVICE_SCSI << 16) + 0x050c



from __future__ import absolute_import, division, print_function, unicode_literals

import sys, os, os.path, logging

from .filesystem import statsislink, isdir, extract_basic_stat_info

def path_split(path):
	head, tail = os.path.split(path)
	root, ext = os.path.splitext(tail)
	return (head, root, ext[1:])

def path_convert_sep(path):
	"""Converts \ to /"""
	return path.replace("\\", "/")

def path_norm(path):
	drive, path = os.path.splitdrive(path)
	return path_convert_sep(os.path.normpath(path))

def path_abs_rel(path, relpath, abs_rel=0):
	"""abs_rel
	0: standard behaviour, abs if directory=abs, rel if directory=rel
	1: abs path
	2: rel path
	3: filename only"""
	if abs_rel == 0:
		return path
	elif abs_rel == 1:
		return os.path.abspath(path)
	elif abs_rel == 2:
		return os.path.relpath(path, relpath)
	elif abs_rel == 3:
		return os.path.basename(path)
	else:
		raise ValueError("abs_rel must be between 0 and 3 inclusive")

#implement more abs,rel options
def listdir_rec(directory, dirs=True, files=True, rec=True, abs_rel=0, link=True, meta=False, onerror=None):
	"""abs_rel
	0: standard behaviour, abs if directory=abs, rel if directory=rel
	1: abs path
	2: rel path
	3: filename only"""
	
	if type(rec) == bool and rec == True:
		rec = -1
	
	try:
		for i in os.listdir(directory):
			next = os.path.join(directory, i)
			stats = os.lstat(next)
			if not isdir(stats):
				if statsislink(stats): # fix 2017-06-19, does not work for junctions...
					print("Skipping link")
					continue
				#((not A) and B) or (A and B and C) == (not A or C) and B
				if files and (not isdir(stats) or link):
					if meta:
						yield (path_abs_rel(next, directory, abs_rel), stats) #path_abs_rel needed?
					else:
						yield path_abs_rel(next, directory, abs_rel) #path_abs_rel needed?
			else:
				if statsislink(stats):
					raise NotImplemented("Directory links are not supported yet")
				if dirs:
					if meta:
						yield (path_abs_rel(next, directory, abs_rel), stats) #path_abs_rel needed?
					else:
						yield path_abs_rel(next, directory, abs_rel) #path_abs_rel needed?
				if rec:
					if meta:
						for sub, substats in listdir_rec(next, dirs, files, rec-1, abs_rel=0, meta=meta, onerror=onerror):
							yield (path_abs_rel(sub, directory, abs_rel), substats) #path_abs_rel needed?
					else:
						for sub in listdir_rec(next, dirs, files, rec-1, abs_rel=0, meta=meta, onerror=onerror):
							yield path_abs_rel(sub, directory, abs_rel)

	except OSError as e:
		if onerror:
			onerror(None, None, sys.exc_info())

def listdir_rec_meta(directory, dirs=True, files=True, rec=True, abs_rel=0):
	for path, stats in listdir_rec(directory, dirs, files, rec, abs_rel, meta=True):
		yield (path, extract_basic_stat_info(stats))

def listdir_rec_adv(directory, dirs=True, files=True, rec=True, total=False):
	try:
		dir_count = 0
		file_count = 0
		for i in os.listdir(directory):
			next = os.path.join(directory, i)
			if not os.path.isdir(next):
				file_count += 1
				if files:
					yield (next, 0, 0)
			else:
				dir_count += 1
				if rec:
					for j, dc, fc in listdir_rec_adv(next, dirs, files, rec, total):
						yield (j, dc, fc)
						if total:
							dir_count += dc
							file_count += fc
		if dirs:
			yield (directory, dir_count, file_count)

	except IOError: pass
	except WindowsError: pass

def folder_contents_rec(directory):
	contents = []
	n = 0
	try:
		for i in os.listdir(directory):
			next = os.path.join(directory, i)
			n += 1
			if os.path.isdir(next):
				contents.extend(folder_contents_rec(next))
		contents.append((directory, n))
	except Exception as e:
		logging.exception("Enumerating directory failed")
		contents.append(("dummy", -1))
	return contents



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems

import logging

import wx
from wx.lib.mixins.listctrl import ColumnSorterMixin

logger = logging.getLogger(__name__)

class AdvancedListCtrl(wx.ListCtrl, ColumnSorterMixin):

	#wx.LC_SORT_* breaks idtopos
	def __init__(self, *args, **kwargs):
		wx.ListCtrl.__init__(self, *args, **kwargs)
		self.columns = 0
		self._init()
		ColumnSorterMixin.__init__(self, 0)

		#self.Bind(wx.EVT_LIST_DELETE_ITEM, self.OnDeleteItem)
		#self.Bind(wx.EVT_LIST_INSERT_ITEM, self.OnInsertItem)

	def _init(self):
		self.rows = 0
		self.itemDataMap = {}
		self.insertcount = 0
		self.idtopos = {}

	def GetListCtrl(self):
		return self

	def ClearAll(self, clear_columns = True):
		if clear_columns:
			wx.ListCtrl.ClearAll(self)
			self.columns = 0
		else:
			wx.ListCtrl.DeleteAllItems(self)
		self._init()

	def Setup(self, columns):
		self.columns = len(columns)
		self.SetColumnCount(self.columns)
		for i, col in enumerate(columns):
			self.InsertColumn(i, col)

	def Append(self, items):
		assert(len(items) == self.columns)
		pos = self.InsertStringItem(self.rows, items[0])
		assert(pos == self.rows)
		#logger.debug("Appending Item on rows->pos: {}->{}".format(self.rows, pos))

		self.rows += 1
		self.SetItemData(pos, self.insertcount)
		self.itemDataMap[self.insertcount] = items

		"""for id, pos in viewitems(self.idtopos):
			if pos >= event.Index:
				self.idtopos[id] = pos + 1"""

		self.idtopos[self.insertcount] = pos
		self.insertcount += 1

		for col in range(1, self.columns):
			self.SetStringItem(pos, col, items[col])

		return self.insertcount - 1

	def Edit(self, pos, items):
		"""items is dict with col as key and string als value"""
		for col, val in viewitems(items):
			self.SetStringItem(pos, col, val)

	def GetSelections(self):
		indexes = []
		pos = self.GetFirstSelected()
		while pos != -1:
			indexes.append(pos)
			pos = self.GetNextSelected(pos)
		return indexes

	def DeleteSelectedItems(self):
		#might be faster with autoarrange (listbox only??) off, no reversed -> autoarrange on
		indexes = self.GetSelections()

		for index in reversed(indexes):
			self.DeleteItem(index)
			self.rows -= 1
			id = self.GetItemData(index)
			del self.itemDataMap[id]
			del self.idtopos[id]

		return indexes

	def Get(self, pos = 0):
		if pos < 0:
			return self.GetItem(self.GetItemCount() + 1 + pos).GetText()
		else:
			return self.GetItem(pos).GetText()

	def IDtoPos(self, id):
		return self.idtopos[id]

	def Delete(self, pos = 0):
		if pos < 0:
			index = self.GetItemCount() + 1 + pos
		else:
			index = pos
		if self.GetItemCount() > pos:
			item = self.GetItem(index)
			ret = item.GetText()
			id = self.GetItemData(index)
			del self.itemDataMap[id]
			del self.idtopos[id]

			self.rows -= 1
			self.DeleteItem(index)
			return ret
		else:
			return None

class AdvancedItemContainerMixin(object):

	def DeleteSelectedItems(self):
		#might be faster with autoarrange of, no reversed -> autoarrange on
		indexes = self.GetSelections()

		for index in reversed(indexes):
			self.Delete(index)

		return indexes

class AdvancedListBox(wx.ListBox, AdvancedItemContainerMixin):

	def __init__(self, *args, **kwargs):
		wx.ListBox.__init__(self, *args, **kwargs)



import numpy as np

class LinearRegression(object):

	""" Linear regression optimized by gradient descent. """

	def __init__(self, X, y, alpha=0.01):
		# type: (float[][], float[], int, float) -> None

		assert len(X.shape) == 2

		self.K = X.shape[1] + 1  # one extra dimension for the bias term
		self.M = len(y)
		self.y = y
		self.α = alpha

		row = np.ones((self.M, 1))
		self.X = np.concatenate([row, X], axis=-1)
		assert self.X.shape == (self.M, self.K)

		self.weights = np.random.random_sample(self.K)

	def predict(self, x):
		# type: (float[], ) -> float

		return np.matmul(x, self.weights)

	def epoch(self):
		# type: () -> None

		self.weights -= self.α * np.matmul(self.X.T, self.predict(self.X) - self.y) # [K]

	def fit(self, n_iter=100, verbose=False):
		# type: () -> None

		for i in range(n_iter):
			if verbose:
				print("SGD itertion:", i)
			self.epoch()

	def getParams(self):
		# type: () -> float[]

		return self.weights[1:]

	def getIntercept(self):
		# type: () -> float

		return self.weights[0]



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, str
from future.moves.urllib import request
from future.moves.urllib.parse import urlparse, urlsplit
from future.moves.urllib.error import URLError

import os, os.path, logging, gzip, errno, json
from io import open
from string import ascii_letters, digits
from typing import TYPE_CHECKING

from .file import copyfilelike, Tell
from .iter import first_not_none
from .filesystem import safe_filename

if TYPE_CHECKING:
	from typing import Callable, Optional, Tuple
	from http.cookiejar import CookieJar

logger = logging.getLogger(__name__)

# URI RFC
gen_delims = ":/?#[]@"
sub_delims = "!$&'()*+;="
reserved = gen_delims + sub_delims
unreserved = ascii_letters + digits + "-._~"
valid_uri_characters = reserved + unreserved

class ContentInvalidLength(Exception):

	def __init__(self, path, expected, received):
		Exception.__init__(self, path, expected, received)

class TimeOut(Exception):
	pass

class DownloadInterrupted(Exception):
	pass

try:
	FileExistsError
except NameError:
	class FileExistsError(IOError):
		pass

try:
	from email.utils import parsedate_to_datetime # new in 3.3

	def parsedate_to_timestamp(datestr): # correct timezone
		return parsedate_to_datetime(datestr).timestamp()

except ImportError:
	import time

	def parsedate_to_timestamp(datestr): # fixme: incorrect timezone
		# type: (str, ) -> float

		try:
			#str fixed py2.x unicode bug
			return time.mktime(time.strptime(datestr, str("%a, %d %b %Y %H:%M:%S %Z")))
		except ValueError:
			#time data 'Thu, 02 Aug 2007 06:54:23 GMT' does not match format u'%a, %d %b %Y %H:%M:%S +0000' # fixed with str()
			return time.mktime(time.strptime(datestr, str("%a, %d %b %Y %H:%M:%S +0000")))

try:
	from email.message import EmailMessage # new in 3.6

	def get_filename(headers):
		# type: (EmailMessage, ) -> str

		return headers.get_filename()

except ImportError:
	def get_filename(headers):
		# type: (Mapping, ) -> str

		try:
			content_disposition = headers["Content-Disposition"]
			return content_disposition.split("=")[1].replace("\"", "").strip()
		except (KeyError, IndexError):
			return None

def get_filename_from_url(url):
	return os.path.split(urlsplit(url).path)[1]

def get_filename_from_url_v2(url):
	return urlparse(url)[2].split("/")[-1]

class URLRequestBuilder(object):

	def __init__(self, cookiejar=None, basicauth=None):
		# type: (Optional[CookieJar], Optional[Tuple[str, str, str]]) -> None

		handlers = []

		if cookiejar:
			handlers.append(request.HTTPCookieProcessor(cookiejar))
		if basicauth:
			top_level_url, username, password = basicauth
			password_mgr = request.HTTPPasswordMgrWithDefaultRealm()
			password_mgr.add_password(None, top_level_url, username, password)
			handlers.append(request.HTTPBasicAuthHandler(password_mgr))

		if handlers:
			opener = request.build_opener(handlers)
			self.openfunc = opener.open
		else:
			self.openfunc = request.urlopen

	def getfunc(self):
		return self.openfunc

	def request(self, url, headers=None, timeout=None):
		return URLRequest(url, headers, timeout, openfunc=self.openfunc)

class FileLike(object):

	def __init__(self, urlrequest):
		self.urlrequest = urlrequest

		self.content_length = self.urlrequest._content_length()

		self.response = Tell(self.urlrequest.response)
		if self.urlrequest.headers.get("Content-Encoding", None) == "gzip":
			# if that does not work, use BytesIO
			logger.info("Using gzip")
			self.fp = gzip.GzipFile(fileobj=self.response)
		else:
			self.fp = self.response

	def close(self):
		self.fp.close()

		transferred = self.response.tell() # is this correct?
		if self.content_length is not None and self.content_length != transferred:
			raise ContentInvalidLength(None, self.content_length, transferred)

	def __enter__(self):
		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		self.close()

class URLRequest(object):

	def __init__(self, url, headers=None, timeout=None, cookiejar=None, basicauth=None, openfunc=None):
		# type: (str, Optional[dict], Optional[float], Optional[CookieJar], Optional[Tuple[str, str, str]], Optional[Callable]) -> None

		self.url = url
		self.timeout = timeout
		headers = headers or {}

		if not openfunc:
			openfunc = URLRequestBuilder(cookiejar=None, basicauth=None).getfunc()

		req = request.Request(url, data=None, headers=headers)
		self.response = openfunc(req, timeout=timeout)
		self.headers = self.response.info()

	def _close(self): # unused
		self.response.close()

	def __enter__(self):
		# type: () -> URLRequest

		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.response.close()

	def _content_length(self):
		# type: () -> Optional[int]

		try:
			content_length = self.headers["Content-Length"]
			return int(content_length)
		except (KeyError, ValueError, TypeError):
			return None

	def _last_modified(self):
		# type: () -> Optional[float]

		try:
			last_modified = self.headers["Last-Modified"]
			return parsedate_to_timestamp(last_modified)
		except (KeyError, TypeError, ValueError):
			return None

	def get_redirect_url(self):
		self.response.close()
		return self.response.geturl()

	def _load(self):
		# type: () -> bytes

		with FileLike(self) as fp:
			try:
				return fp.read()
			except URLError:
				raise TimeOut("Timed out after {}s".format(self.timeout))

	def _json(self):
		# type: () -> dict

		with FileLike(self) as fp:
			try:
				return json.load(fp)
			except URLError:
				raise TimeOut("Timed out after {}s".format(self.timeout))

	def _download(self, basepath, filename=None, fn_prio=None, overwrite=False, suffix=".partial", report=None):
		# type: (str, Optional[str], Optional[Tuple[int, int, int, int]], bool, str, Optional[Callable[[int, int], None]]) -> Tuple[Optional[int], str]

		"""PermissionError: [WinError 32] Der Prozess kann nicht auf die Datei zugreifen,
			da sie von einem anderen Prozess verwendet wird:
			'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a.partial'
			-> 'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a'
		"""

		logger.debug("Downloading %s to %s", self.url, basepath)

		if fn_prio is None:
			fn_prio = (0, 1, 2, 3)

		filenames = {
			0: filename,
			1: get_filename(self.headers),
			2: get_filename_from_url(self.response.geturl()), # url after redirect
			3: get_filename_from_url(self.url),
		}

		filenames = [filenames[p] for p in fn_prio]
		logger.info("Filenames: {}".format(", ".join(map(str, filenames))))

		filename = first_not_none(filenames)

		if not filename:
			raise ValueError("Please provide a filename")

		filename = safe_filename(filename)
		fullpath = os.path.join(basepath, filename)
		tmppath = fullpath + suffix

		if not overwrite and os.path.exists(fullpath):
			raise FileExistsError(errno.EEXIST, "File already exists", fullpath)

		try:
			os.makedirs(basepath) # exist_ok=True is new in 3.2
		except OSError:
			pass

		content_length = self._content_length()

		try:
			# actually download the file
			with open(tmppath, "wb") as out:
				# https://www.ietf.org/mail-archive/web/httpbisa/current/msg27484.html
				transferred = copyfilelike(self.response, out, content_length, report=report)

		except URLError:
			logger.warning("Timeout after {}s at {}: {}".format(self.timeout, self.response.geturl(), self.headers))
			raise TimeOut("Timed out after {}s".format(self.timeout))

		except ConnectionResetError as e:
			logger.warning("Connection was reset during download: %s", str(e))
			raise DownloadInterrupted("Connection was reset during download")

		except Exception as e:
			logger.exception("Downloading failed mid file. Handle.")
			raise

		if content_length and content_length != transferred:
			logger.info("{} {}".format(content_length, transferred))
			raise ContentInvalidLength(tmppath, content_length, transferred)

		last_modified = self._last_modified()
		if last_modified:
			os.utime(tmppath, (-1, last_modified))

		os.rename(tmppath, fullpath)

		return (content_length, filename)

	def load(self):
		# type: () -> bytes

		try:
			return self._load()
		finally:
			self.response.close()

	def json(self):
		# type: () -> bytes

		try:
			return self._json()
		finally:
			self.response.close()

	def download(self, basepath, filename=None, fn_prio=None, overwrite=False, suffix=".partial", report=None):
		# type: () -> Tuple[Optional[int], str]

		try:
			return self._download(basepath, filename, fn_prio, overwrite, suffix, report)
		finally:
			self.response.close()

if __name__ == "__main__":
	from argparse import ArgumentParser

	parser = ArgumentParser(description="Downloads url to file")
	parser.add_argument("url", help="url to download")
	parser.add_argument("dir", help="output directory")
	parser.add_argument("-f", "--filename", help="filename")
	args = parser.parse_args()

	URLRequest(args.url).download(args.dir, args.filename)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import PY2
from builtins import str, bytes

if PY2:

	import io, os
	from gzip import GzipFile
	from pathlib2 import PurePath

	os.PathLike = PurePath

	""" This function is copied from `Python36\Lib\gzip.py`.
	"""

	def open(filename, mode="rb", compresslevel=9,
			 encoding=None, errors=None, newline=None):
		"""Open a gzip-compressed file in binary or text mode.

		The filename argument can be an actual filename (a str or bytes object), or
		an existing file object to read from or write to.

		The mode argument can be "r", "rb", "w", "wb", "x", "xb", "a" or "ab" for
		binary mode, or "rt", "wt", "xt" or "at" for text mode. The default mode is
		"rb", and the default compresslevel is 9.

		For binary mode, this function is equivalent to the GzipFile constructor:
		GzipFile(filename, mode, compresslevel). In this case, the encoding, errors
		and newline arguments must not be provided.

		For text mode, a GzipFile object is created, and wrapped in an
		io.TextIOWrapper instance with the specified encoding, error handling
		behavior, and line ending(s).

		"""
		if "t" in mode:
			if "b" in mode:
				raise ValueError("Invalid mode: %r" % (mode,))
		else:
			if encoding is not None:
				raise ValueError("Argument 'encoding' not supported in binary mode")
			if errors is not None:
				raise ValueError("Argument 'errors' not supported in binary mode")
			if newline is not None:
				raise ValueError("Argument 'newline' not supported in binary mode")

		gz_mode = mode.replace("t", "")
		if isinstance(filename, (str, bytes, os.PathLike)):
			binary_file = GzipFile(filename, gz_mode, compresslevel)
		elif hasattr(filename, "read") or hasattr(filename, "write"):
			binary_file = GzipFile(None, gz_mode, compresslevel, filename)
		else:
			raise TypeError("filename must be a str or bytes object, or a file")

		if "t" in mode:
			return io.TextIOWrapper(binary_file, encoding, errors, newline)
		else:
			return binary_file

else:
	from gzip import open



from __future__ import absolute_import, division, print_function, unicode_literals

from io import open

from nltk.tokenize import word_tokenize

def count_words_in_file(path, encoding="utf-8"):
	# type: (Path, ) -> int

	with open(path, "r", encoding=encoding) as fr:
		return sum(len(word_tokenize(line)) for line in fr)



from __future__ import absolute_import, division, print_function, unicode_literals

from io import open
import csv
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Dict, Iterator

def parse_imdb_csv(path):
	# type: (str, ) -> Iterator[Dict[str, str]]

	with open(path, "r", encoding="iso-8859-1", errors="strict", newline="") as fr:
		csvreader = csv.DictReader(fr)
		for row in csvreader:
			yield row



from __future__ import absolute_import, division, print_function, unicode_literals

DAR_WS = 16/9
DAR_FS = 4/3

RES_PAL = (720, 576)
RES_NTSC = (720, 480)
RES_PAL_BLANKING = (704, 576)
RES_NTSC_BLANKING = (704, 480)
RES_PAL_HALF_H = (352, 576)
RES_NTSC_HALF_H = (352, 480)
RES_HALF_HD = (1280, 720)
RES_FULL_HD = (1920, 1080)

PAR_HD = 1

def SAR(RES):
	x, y = RES
	return x/y

#Generic / Digital PAR
PAR_PAL_16_9 = DAR_WS/SAR(RES_PAL) #64:45
PAR_PAL_4_3 = DAR_FS/SAR(RES_PAL) #16:15
PAR_PAL_4_3_BLANKING = DAR_FS/SAR(RES_PAL_BLANKING) #12:11
PAR_PAL_HALF_H = DAR_FS/SAR(RES_PAL_HALF_H) #24:11
PAR_NTSC_16_9 = DAR_WS/SAR(RES_NTSC) #32:27
PAR_NTSC_4_3 = DAR_FS/SAR(RES_NTSC) #8:9
PAR_NTSC_4_3_BLANKING = DAR_FS/SAR(RES_NTSC_BLANKING) #10:11
PAR_NTSC_HALF_H = DAR_FS/SAR(RES_NTSC_HALF_H) #20:11

#MPEG-4 PAR
PAR_MPG4_PAL_16_9 = PAR_PAL_OVERSCAN_16_9 = DAR_WS/SAR(RES_PAL_BLANKING) #16:11
PAR_MPG4_PAL_4_3 = PAR_PAL_OVERSCAN_4_3 = DAR_FS/SAR(RES_PAL_BLANKING) #12:11
PAR_MPG4_NTSC_16_9 = PAR_NTSC_OVERSCAN_16_9 = DAR_WS/SAR(RES_NTSC_BLANKING) #40:33
PAR_MPG4_NTSC_4_3 = PAR_NTSC_OVERSCAN_4_3 = DAR_FS/SAR(RES_NTSC_BLANKING) #10:11

#ITU-R BT.601 PAR (Rec. 601)
PAR_ITU_PAL_16_9 = 512/351
PAR_ITU_PAL_4_3 = 128/117
PAR_ITU_NTSC_16_9 = 5760/4739
PAR_ITU_NTSC_4_3 = 4320/4739

#Frames per seconds
REAL_FILM_FPS = 24
NTSC_PROG_FPS = 24000/1001
NTSC_VIDEO_FPS = 30000/1001
PAL_PROG_FPS = 25
PAL_VIDEO_FPS = 30



from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from scipy.optimize import linear_sum_assignment
from scipy.linalg import solve_triangular

# from .numba import opjit

def linear_sum_assignment_cost(cost_matrix):
	""" Minimum weight matching in bipartite graphs using the Hungarian/Kuhn-Munkres algorithm. """

	row_ind, col_ind = linear_sum_assignment(cost_matrix)
	return np.sum(cost_matrix[row_ind, col_ind])

#@opjit(["f4[:,:](f4[:,:,:], f4[:,:], b1, b1, b1)", "f8[:,:](f8[:,:,:], f8[:,:], b1, b1, b1)"])
def batch_solve_triangular(matrices, vectors, lower=False, overwrite_b=False, check_finite=False):
	# type: (np.ndarray, np.ndarry, bool, bool) -> np.ndarray

	B, M1, M2 = matrices.shape
	B2, V = vectors.shape

	if B != B2 or M1 != M2 or M2 != V:
		raise ValueError("Input dimensions are invalid")

	out = np.empty_like(vectors)

	for i in range(B):
		out[i] = solve_triangular(matrices[i], vectors[i], lower=lower, overwrite_b=overwrite_b, check_finite=check_finite)

	return out

if __name__ == "__main__":
	import timeit
	from .numpy import random_triangular_matrix

	size = 100
	bs = 200
	Ms = np.array([random_triangular_matrix(size, lower=True) for _ in range(bs)])
	vs = np.random.uniform(0, 1, (bs, size))

	batch_solve_triangular(Ms, vs, lower=True) # warmup
	print(min(timeit.repeat('batch_solve_triangular(Ms, vs, lower=True)', number=10, repeat=5, globals=globals())))



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from itertools import product
from operator import itemgetter
from typing import TYPE_CHECKING

from .func import identity

if TYPE_CHECKING:
	from typing import Callable, Iterable, Iterator, MutableSequence, Tuple

class LasyStringList(object):

	def __init__(self, length):
		self.length = length

	def __getitem__(self, idx):
		return str(idx)

	def __len__(self):
		return self.length

	def __iter__(self):
		return map(str, range(self.length))

	def __contains__(self, idx):
		return 0 <= idx < self.length

def triangular(seq):
	# type: (Sequence[T], ) -> Iterator[Tuple[T, T]]

	""" Returns all combinations of items of `seq` with duplicates and self-combinations.
		triangular([1, 2, 3, 4]) -> (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)
	"""

	for i in range(len(seq)-1):
		for j in range(i+1, len(seq)):
			yield seq[i], seq[j]

def sliding_window(seq, size):
	# type: (Sequence, int) -> Iterator[Sequence]

	seqlen = len(seq)
	assert size <= seqlen

	for i in range(0, seqlen-size+1):
		yield seq[i:i+size]

# was: read_x
def batch(seq, size):
	# type: (Sequence, int) -> Iterator[Sequence]

	seqlen = len(seq)

	for i in range(0, (seqlen + size - 1) // size):
		yield seq[i*size:(i+1)*size]

def merge(lists, optimize_for_speed=True):
	# type: (Iterable[Iterable], ) -> list

	""" merge lists into one, while removing duplicates but keeping element order.
		`optimize_for_speed` is True has better runtime complexity
		`optimize_for_speed` is False uses less memory
	"""

	merged = []

	if optimize_for_speed:
		uniques = set()
		for list in lists:
			for element in list:
				if element not in uniques:
					merged.append(element)
					uniques.add(element)

	else:
		for list in lists:
			for element in list:
				if element not in merged:
					merged.append(element)

	return merged

def delete_duplicates_from_sorted_sequence(seq, key=None):
	# type: (MutableSequence, Callable) -> None

	""" Deletes duplicates from a sorted sequence `seq` based on `key`.
		Works in-place.
	"""

	key = key or identity
	i = 0
	while i < len(seq) - 1:
		if key(seq[i]) == key(seq[i+1]):
			del seq[i+1]
		else:
			i += 1

remove_duplicate_rows_from_sorted_table = lambda seq, key: delete_duplicates_from_sorted_sequence(seq, itemgetter(key))

def cycle_sort(sequence):
	# type: (MutableSequence, ) -> Iterator[Tuple[int, int]]

	""" Sort a sequence in place and yield swaps.
		based on: https://rosettacode.org/wiki/Sorting_algorithms/Cycle_sort#Python
	"""

	# Loop through the sequence to find cycles to rotate.
	for cycle_start, item in enumerate(sequence):

		# Find where to put the item.
		pos = cycle_start
		for item2 in sequence[cycle_start + 1:]:
			if item2 < item:
				pos += 1

		# If the item is already there, this is not a cycle.
		if pos == cycle_start:
			continue

		# Otherwise, put the item there or right after any duplicates.
		while item == sequence[pos]:
			pos += 1
		sequence[pos], item = item, sequence[pos]
		yield cycle_start, pos

		# Rotate the rest of the cycle.
		while pos != cycle_start:

			# Find where to put the item.
			pos = cycle_start
			for item2 in sequence[cycle_start + 1:]:
				if item2 < item:
					pos += 1

			# Put the item there or right after any duplicates.
			while item == sequence[pos]:
				pos += 1
			sequence[pos], item = item, sequence[pos]
			if pos != cycle_start:
				yield cycle_start, pos



from __future__ import absolute_import, division, print_function, unicode_literals

import csv
from itertools import islice
from functools import partial

from .dict import itemgetter
from .func import identity, zipmap, compose

def read_csv(path, delimiter=",", skip=0, usecols=None, dtype=None):
	with open(path, encoding="utf-8", newline="") as csvfile:
		if usecols:
			if dtype:
				getcols = compose(partial(zipmap, dtype), itemgetter(usecols))
			else:
				getcols = itemgetter(usecols)
		else:
			if dtype:
				getcols = partial(zipmap, dtype)
			else:
				getcols = identity

		for row in islice(csv.reader(csvfile, delimiter=delimiter), skip, None):
			print(row)
			yield tuple(getcols(row))



from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

from .numpy import unblock

def grayscale(arr):
	return np.sum(arr, axis=-1) // arr.shape[-1]

def histogram_1d(arr, levels):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x]. Histogrammed over x and batched over the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	return np.apply_along_axis(np.bincount, -1, arr, minlength=levels)

def histogram_2d(arr, levels):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y]. Histogrammed over x and y and batched over
		the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	assert len(arr.shape) >= 2

	newshape = arr.shape[:-2] + (arr.shape[-2] * arr.shape[-1], )
	flattened = np.reshape(arr, newshape)
	return np.apply_along_axis(np.bincount, -1, flattened, minlength=levels)

def block_histogram_2d(arr, bx, by, levels):
	# type: (np.ndarray, int, int, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y]. Histogrammed over x and y and batched over
		the remaining dimensions.
		Output shape: [batch..., bx, by, levels]
	"""

	invx = arr.shape[-1] // bx  # dimensions in unblock go from innerst to outerst
	invy = arr.shape[-2] // by
	blocks = unblock(arr, invx, invy)
	block_hists = histogram_1d(blocks, levels)
	return block_hists.reshape(arr.shape[:-2] + (invy, invx, -1))

def image_histogram(arr, levels=256):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y, channel]. It is summed over channels to create a grayscale
		image, then histogrammed over x and y and batched over the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	assert len(arr.shape) >= 3

	gray = grayscale(arr)
	return histogram_2d(gray, levels)

def image_block_histogram(arr, bx, by, levels=256):
	# type: (np.ndarray, int, int, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y, channels]. It is summed over channels to create a grayscale
		image, then histogrammed over x and y and batched over the remaining dimensions.
		Output shape: [batch..., bx, by, levels]
	"""

	assert len(arr.shape) >= 3

	gray = grayscale(arr)
	return block_histogram_2d(gray, bx, by, levels)



from __future__ import absolute_import, division, print_function, unicode_literals

import sqlite3

def sqlite_vacuum(db_path):
	# type: (str, ) -> None

	""" Vacuums a sqlite database. """

	with sqlite3.connect(db_path) as conn:
		conn.execute("VACUUM")



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime as _datetime

try:
	from datetime import timezone

except ImportError:

	from datetime import timedelta, tzinfo

	class timezone(tzinfo):

		def __init__(self, offset, name=None):
			# type: (int, Optional[str]) -> None

			self.offset = offset
			self.name = name

		def __reduce__(self):
			return (timezone, (self.offset, self.name))

		def utcoffset(self, dt):
			return self.offset

		def dst(self, dt):
			return timedelta(0)

		def tzname(self, dt):
			if self.name:
				return self.name
			else:
				raise NotImplementedError

	timezone.utc = timezone(timedelta(0))

try:
	_datetime.fromisoformat
	datetime = _datetime

except AttributeError:

	import re
	from datetime import timedelta
	from ..string import tryint

	isoformatre = re.compile(r"^(\d{4})-(\d{2})-(\d{2})[ T](\d{2}):(\d{2}):(\d{2})\.(\d{6})(\+|-)(\d{2}):(\d{2})$")

	class datetime(_datetime):

		@classmethod
		def fromisoformat(cls, date_string):
			# (str, ) -> datetime

			""" Converts an ISO 8601 date and time with timezone string to a timezone aware datetime.
				eg. '2019-03-01T11:30:01.123000+00:00'
				eg. '2019-02-02T12:30:02.456000+04:00'
				eg. '2019-01-03T22:30:03.789000-07:00'
			"""

			groups = isoformatre.match(date_string).groups()
			year, month, day, hour, minute, second, milliseconds, tzsign, tzhours, tzminutes = map(tryint, groups)
			offset = timedelta(hours=tzhours, minutes=tzminutes)

			if tzsign == "-":
				offset = -offset

			return cls(year, month, day, hour, minute, second, milliseconds, tzinfo=timezone(offset))



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from io import open

from ..exceptions import MalformedFile

REAL_FILM_FPS = 24.
NTSC_PROG_FPS = 24000./1001
NTSC_VIDEO_FPS = 30000./1001
PAL_PROG_FPS = 25.
PAL_VIDEO_FPS = 30.

def ntsc_to_pal(num):
	return num * NTSC_PROG_FPS/PAL_PROG_FPS

def pal_to_ntsc(num):
	return num * PAL_PROG_FPS/NTSC_PROG_FPS

def film_to_ntsc(num):
	return num * REAL_FILM_FPS/NTSC_PROG_FPS

def to_msec(string):
	h, m, s, ms = [int(i) for i in string.replace(",", ":").replace(".", ":").split(":")]
	return (h * 60 * 60 + m * 60 + s)*1000 + ms

def to_srt_time(t):
	r, ms = divmod(t, 1000)
	r, s = divmod(r, 60)
	h, m = divmod(r, 60)
	return "{:02d}:{:02d}:{:02d},{:03d}".format(int(h), int(m), int(s), int(ms))

class Subtitle(object):

	def __init__(self):
		self.num = 0
		self.start = 0
		self.end = 0
		self.lines = []

	def get_times(self):
		return to_srt_time(self.start), to_srt_time(self.end)

	def set_times(self, start, end):
		self.start, self.end = to_msec(start), to_msec(end)

	def append(self, line):
		self.lines.append(line)

class SRTFile(object):

	nl = "\n"
	sep = " --> "

	def __init__(self, filename, mode="r", encoding="utf-8-sig", overwrite_index=False):
		self.state = 0
		self.linenum = 0
		self.sub_num = 0
		self.fp = open(filename, mode, encoding=encoding, errors="replace")
		self.overwrite_index = overwrite_index

	def __enter__(self):
		return self

	def __exit__(self, type, value, traceback):
		self.close()

	def close(self):
		self.fp.close()

	def write_subtitle(self, subtitle):

		start, end = subtitle.get_times()
		if self.overwrite_index:
			self.fp.write(str(self.sub_num) + self.nl)
		else:
			self.fp.write(str(subtitle.num) + self.nl)
		self.fp.write(start + self.sep + end + self.nl)
		for line in subtitle.lines:
			self.fp.write(line)
			self.fp.write(self.nl)
		self.fp.write(self.nl)
		self.sub_num += 1

	def read_line(self):
		self.linenum += 1
		line = next(self.fp).rstrip()

		if self.state == 0:
			try:
				self.sub.num = int(line)
			except ValueError:
				raise MalformedFile("Error in line {}: srt malformed: {}".format(self.linenum,line))
			self.state = 1
		elif self.state == 1:
			start, end = line.split(self.sep)
			self.sub.set_times(start, end)
			self.state = 2
		elif self.state == 2:
			if line == "":
				self.state = 0
			else:
				self.sub.append(line)

	def read_subtitle(self):
		self.sub = Subtitle()
		self.read_line()
		while self.state != 0:
			self.read_line() # can throw, and leave incomplete subtitle in buffer (eg. if file doesnt end in newline)
		return self.sub

	def __iter__(self):
		if self.fp:
			return self
		else:
			raise ValueError("I/O operation on closed file.")

	def __next__(self):
		return self.read_subtitle()

def transform(infile, outfile, callback, encoding):
	with SRTFile(infile, "r", encoding) as fi, SRTFile(outfile, "w", encoding) as fo:
		for subtitle in fi:
			subtitle = callback(subtitle)
			fo.write_subtitle(subtitle)

def merge(srtlines, lim):
	ret = []
	for l in srtlines:
		if l.startswith(lim):
			ret.append(l[len(lim):])
		else:
			if not ret:
				ret.append(l)
			else:
				ret[-1] = ret[-1] + " " + l
	return ret

def srt2txt(srt_fp):

	lim = "- "
	pos = 0
	lines = []

	while True:
		if pos >= len(lines):
			sub = next(srt_fp)
			lines = merge(sub.lines, lim)
			pos = 0
		ret = lines[pos]
		pos += 1
		yield ret

def compare_srt_and_txt(srt_file, txt_file):

	with SRTFile(srt_file, "r", encoding="utf-8-sig") as srt, open(txt_file, "r", encoding="utf-8-sig") as txt:

		srtiter = srt2txt(srt)

		limit = 10

		while limit > 0:
			try:
				l = next(txt).rstrip()
				if not l:
					continue
				if not l.startswith("- "):
					print(l)
				else:
					l = l[2:]
					start = 0
					while start <= len(l):
						srtpart = next(srtiter)
						txtpart = l[start:start+len(srtpart)]
						logging.debug("Compare '{}' '{}'".format(txtpart, srtpart))
						if txtpart != srtpart:
							#print("{} - {}".format(sub.start, sub.end))
							yield (txtpart, srtpart)
							limit -= 1
						start += len(srtpart) + 1
			except StopIteration:
				break



from twisted.internet import protocol

class GenericFactory(protocol.Factory):
	def __init__(self, protocol, *args, **kwargs):
		self.protocol = protocol
		self.args = args
		self.kwargs = kwargs

	def buildProtocol(self, addr):
		return self.protocol(self, *self.args, **self.kwargs) #deleted "self, " here...



from __future__ import absolute_import, division, print_function, unicode_literals

import sys

from future.utils import PY2

if PY2:
	def tofs(s):
		return s.encode(sys.getfilesystemencoding())

	def fromfs(b):
		return b.decode(sys.getfilesystemencoding())

	def sbs(s):
		return s.encode("ascii")

else:
	def tofs(s):
		return s

	def fromfs(b):
		return b

	def sbs(s):
		return s



from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib, zlib, os.path
from functools import partial
from typing import TYPE_CHECKING

from .file import read_file, blockfileiter
from .filesystem import scandir_rec

if TYPE_CHECKING:
	from typing import Callable, Optional
	from _hashlib import HASH as Hashobj

FILE_IO_BUFFER_SIZE = 8*1024*1024

def hash_file(path, hashcls, chunksize=FILE_IO_BUFFER_SIZE, mode="rb", encoding=None):
	# type: (str, Callable[[], Hashobj], int, str, Optional[str]) -> Hashobj

	m = hashcls()
	for d in blockfileiter(path, mode, encoding, chunk_size=chunksize):
		m.update(d)
	return m

def crc32_hash_iter(it):
	# type: (Iterable[bytes], ) -> int

	prev = 0
	for data in it:
		prev = zlib.crc32(data, prev)

	return prev & 0xFFFFFFFF # see https://docs.python.org/3/library/zlib.html#zlib.crc32

def crc32_hash_file(path, chunksize=FILE_IO_BUFFER_SIZE, mode="rb", encoding=None):
	# type: (str, int, str, Optional[str]) -> str

	""" Return crc32 hash of file at `path`. """

	crcint = crc32_hash_iter(blockfileiter(path, mode, encoding, chunk_size=chunksize))
	return format(crcint, "08x")

md5_hash_file = partial(hash_file, hashcls=hashlib.md5)
sha1_hash_file = partial(hash_file, hashcls=hashlib.sha1)

def format_file_hash(hashobj, path):
	# type: (Hashobj, str) -> str

	return "{} *{}".format(hashobj.hexdigest(), path)

def hash_dir_str(path, hashcls=hashlib.sha1, include_names=False):
	# type: (str, Callable[[], Hashobj], bool) -> str

	""" sorts names naively, e.g. all uppercase chars come before lowercase """

	m = hashcls()
	for file in sorted(scandir_rec(path), key=lambda x: x.path):
		filehash = hash_file(file, hashcls)
		yield format_file_hash(filehash, file.path)
		if include_names:
			m.update(file.name.encode("utf-8"))
		m.update(filehash.digest())
	yield format_file_hash(m, path)

def md4_hash_data(data):
	# type: (bytes,) -> Hashobj

	hasher = hashlib.new("md4")
	hasher.update(data)
	return hasher

ed2k_chunksize = 9728000

def ed2k_hash_file_v1(path):
	""" Returns ed2k hash.
		This hashing method is used by
		- MLDonkey
		- Shareaza
		- HashCalc
	"""

	if os.path.getsize(path) <= ed2k_chunksize:
		return md4_hash_data(read_file(path, "rb")).hexdigest()

	ed2k_hashes = (
		md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize)
	)

	return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()

def ed2k_hash_file_v2(path):
	""" Returns ed2k hash.
		This hashing method is used by
		- eMule
		- AOM
	"""

	filesize = os.path.getsize(path)
	if filesize < ed2k_chunksize:
		return md4_hash_data(read_file(path, "rb")).hexdigest()

	ed2k_hashes = list(
		md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize)
	)

	if filesize % ed2k_chunksize == 0:
		ed2k_hashes.append(md4_hash_data(b"").digest())

	return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()



from __future__ import absolute_import, division, print_function, unicode_literals

from functools import reduce

import rhash

from .file import blockfileiter
from .ops import bit_or
from .hash import FILE_IO_BUFFER_SIZE

def multi_hash_file(path, hash_types, base=None, mode="rb", encoding=None, errors=None, chunk_size=FILE_IO_BUFFER_SIZE):

	""" Hashes `path` with multiple hash functions.
		`hash_types` is a list of names, eg. ("CRC32", "SHA1").
	"""

	hashdict = {
		"CRC32": rhash.CRC32,
		"MD4": rhash.MD4,
		"MD5": rhash.MD5,
		"SHA1": rhash.SHA1,
		"TIGER": rhash.TIGER,
		"TTH": rhash.TTH,
		"BTIH": rhash.BTIH,
		"ED2K": rhash.ED2K,
		"AICH": rhash.AICH,
		"WHIRLPOOL": rhash.WHIRLPOOL,
		"RIPEMD-160": rhash.RIPEMD160,
		"GOST12_256": rhash.GOST12_256,
		"GOST12_512": rhash.GOST12_512,
		"GOST94": rhash.GOST94,
		"GOST94_CRYPTOPRO": rhash.GOST94_CRYPTOPRO,
		"HAS-160": rhash.HAS160,
		"SNEFRU-128": rhash.SNEFRU128,
		"SNEFRU-256": rhash.SNEFRU256,
		"SHA-224": rhash.SHA224,
		"SHA-256": rhash.SHA256,
		"SHA-384": rhash.SHA384,
		"SHA-512": rhash.SHA512,
		"EDON-R256": rhash.EDONR256,
		"EDON-R512": rhash.EDONR512
	}

	basedict = {
		"upper": rhash.RHPR_UPPERCASE,
		"raw": rhash.RHPR_RAW,
		"hex": rhash.RHPR_HEX,
		"HEX": rhash.RHPR_HEX|rhash.RHPR_UPPERCASE,
		"base32": rhash.RHPR_BASE32,
		"BASE32": rhash.RHPR_BASE32|rhash.RHPR_UPPERCASE,
		"base64": rhash.RHPR_BASE64,
	}

	hashlist = tuple(hashdict[i.upper()] for i in hash_types)
	flags = basedict.get(base, 0)

	hasher = rhash.RHash(reduce(bit_or, hashlist))
	for data in blockfileiter(path, mode, encoding, errors, chunk_size=chunk_size):
		hasher.update(data)
	hasher.finish()
	return tuple(hasher._print(i, flags) for i in hashlist)

if __name__ == "__main__":
	import argparse

	funcs = (
		"CRC32", "MD4", "MD5", "SHA1", "TIGER", "TTH", "BTIH", "ED2K", "AICH", "WHIRLPOOL",
		"RIPEMD-160", "GOST12_256", "GOST12_512", "GOST94", "GOST94_CRYPTOPRO", "HAS-160", "SNEFRU-128", "SNEFRU-256", "SHA-224",
		"SHA-256", "SHA-384", "SHA-512", "EDON-R256", "EDON-R512"
	)

	parser = argparse.ArgumentParser(description="Multi file hashing", epilog="Example: multihash C:\\Windows\\System32\\calc.exe -f CRC32 SHA1")
	parser.add_argument("-f", "--hashfuncs", nargs='+', help='list of hash functions. Can be one or more of: {}'.format(", ".join(funcs)), required=True)
	parser.add_argument("file", metavar="FILE", help="path to the file")
	args = parser.parse_args()
	print("\n".join("{}: {}".format(i, j) for i, j in zip(args.hashfuncs, multi_hash_file(args.file, args.hashfuncs))))



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

if TYPE_CHECKING:

	from typing import Any
	from typing_extensions import Protocol

	class Computable(Protocol):

		def __neg__(self, other):
			# type: (Any, ) -> Any
			...

		def __add__(self, other):
			# type: (Any, ) -> Any
			...

		def __sub__(self, other):
			# type: (Any, ) -> Any
			...

		def __mul__(self, other):
			# type: (Any, ) -> Any
			...

		def __truediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __floordiv__(self, other):
			# type: (Any, ) -> Any
			...

		def __abs__(self, other):
			# type: (Any, ) -> Any
			...

	class MutableComputable(Protocol):

		def __neg__(self, other):
			# type: (Any, ) -> Any
			...

		def __add__(self, other):
			# type: (Any, ) -> Any
			...

		def __iadd__(self, other):
			# type: (Any, ) -> Any
			...

		def __sub__(self, other):
			# type: (Any, ) -> Any
			...

		def __isub__(self, other):
			# type: (Any, ) -> Any
			...

		def __mul__(self, other):
			# type: (Any, ) -> Any
			...

		def __imul__(self, other):
			# type: (Any, ) -> Any
			...

		def __truediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __itruediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __floordiv__(self, other):
			# type: (Any, ) -> Any
			...

		def __ifloordiv__(self, other):
			# type: (Any, ) -> Any
			...



from __future__ import absolute_import, division, print_function, unicode_literals

import re

from .http import valid_uri_characters

def extract_urls(text):
	pattern = r"https?\:\/\/[" + re.escape(valid_uri_characters) + r"]+"
	for match in re.finditer(pattern, text):
		url = match.group(0)
		yield url



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from genutility.math import argmin, argmax

if TYPE_CHECKING:
	from typing import MutableSequence

def bubble_sort(seq):
	# type: (MutableSequence, ) -> None

	""" Slightly optimized version of stable bubble sort. Sorts input `seq` in place. """

	n = len(seq)
	while True:
		newn = 1
		for i in range(1, n):
			if seq[i-1] > seq[i]:
				seq[i-1], seq[i] = seq[i], seq[i-1]
				newn = i
		if newn == 1:
			break
		n = newn

def selection_sort_min(l):
	#inplace, unstable

	n = len(l)
	for i in range(n-1):
		m = argmin(l, i, n)
		if m != i:
			l[i], l[m] = l[m], l[i]

def selection_sort_max(l):
	#inplace, unstable

	for n in range(len(l)-1, 0, -1):
		m = argmax(l[:n+1])
		if m != n:
			l[m], l[n] = l[n], l[m]

def selection_sort_ll(l):
	#inplace, stable, on linked lists
	raise NotImplementedError



from __future__ import absolute_import, division, print_function, unicode_literals

STARS = {
	"Sun": {
		"mass": 1.9891e30,
	}
}

PLANETS = {
	"Earth": {
		"mass": 5.9736e24,
	}
}

UNITS = {
	"Astronomical_unit": 149597870700, # m
	"Day": 60*60*24, # s
}

FUNDAMENTAL = {
	"Gravitational_constant": 6.67384e-11,
}



from __future__ import absolute_import, division, print_function, unicode_literals

from setuptools import setup
from io import open

# nltk.download("punkt")

with open("README.md", "r", encoding="utf-8") as fr:
	long_description = fr.read()

setup(
	author="Dobatymo",
	name="genutility",
	version="0.0.11",
	url="https://github.com/Dobatymo/genutility",
	description="A collection of various Python utilities",
	long_description=long_description,
	long_description_content_type="text/markdown",
	classifiers=[
		"Programming Language :: Python :: 2",
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: ISC License (ISCL)",
		"Operating System :: OS Independent",
		"Topic :: Utilities",
	],
	packages=["genutility", "genutility/compat", "genutility/fileformats", "genutility/hardware", "genutility/twothree"],
	python_requires=">=2.7",
	install_requires=[
		"future",
		"typing;python_version<'3.5'",
		"scandir;python_version<'3.5'",
		"pathlib;python_version<'3.4'",
		"mock;python_version<'3.3'",
		"contextlib2;python_version<'3.3'",
	],
	extras_require={
		"gensim": ["gensim>=3.3.0", "numpy"],
		"pdf": ["PyPDF2"],
		"tls": ["pycryptodome", "pyOpenSSL"],
		"toml": ["toml"],
		"ALL": ["aiohttp", "bencode.py", "ctypes-windows-sdk", "flask", "gensim", "msgpack",
			"netifaces", "nltk", "orderedset", "pycryptodome", "pyOpenSSL", "PyPDF2", "pypiwin32",
			"rhash", "toml", "unidecode", "werkzeug", "wmi", "wx"],
	},
	use_2to3=False
)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from future.utils import viewitems

from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

from .numba import opjit

if TYPE_CHECKING:
	from typing import Iterator, Tuple

RGB_YELLOW = (255, 255, 0)
RGB_WHITE = (255, 255, 255)

def normalized_choice(p_ind, p_prob):  # is this neccessary? I think `np.random.choice` can handle unnormalized probabilities
	return np.random.choice(p_ind, p=p_prob/np.sum(p_prob))

def is_rgb(img):
	# type: (np.ndarray, ) -> bool

	return len(img.shape) >= 1 and img.shape[-1] == 3

#@opjit() np.errstate doesn't work in numba...
def rgb_to_hsi(image):
	# type: (np.ndarray, ) -> np.ndarray

	""" Converts an array [..., channels] of RGB values to HSI color values (H in rad).
		RGB values are assumed to be normalized to (0, 1).
	"""

	if not is_rgb(image):
		raise ValueError("Input needs to be an array of RGB values")

	r = image[...,0]
	g = image[...,1]
	b = image[...,2]

	out = np.zeros_like(image)

	#allequal = (img == img[:, :, 0, np.newaxis]).all(axis=-1)

	with np.errstate(invalid="ignore"): 
		tmp = (2.*r - g - b) / 2. / np.sqrt((r-g)**2 + (r-b)*(g-b)) # if r==g==b then 0/0

		theta = np.arccos(np.clip(tmp, -1., +1.))

		out[...,0] = np.where(b <= g, theta, 2*np.pi - theta) # H
		out[...,2] = np.sum(image, axis=-1) / 3. # I
		out[...,1] = 1 - np.amin(image, axis=-1) / out[...,2] # S if r==g==b==0 then 0/0

	np.nan_to_num(out[...,0:2], copy=False)

	return out

#@opjit() np.dot with more than 2 dimensions is not supported...
def rgb_to_ycbcr(image):
	# type: (np.ndarray, ) -> np.ndarray

	""" Converts an array [..., channels] of RGB values to Digital Y'CbCr (0-255).
		RGB values are assumed to be normalized to (0, 1).
		Don't forget to cast to uint8 for pillow.
	"""


	"""  from RGB (0-1).
	"""

	if not is_rgb(image):
		raise ValueError("Input needs to be an array of RGB values")

	m = np.array([
		[+065.481, +128.553, +024.966],
		[-037.797, -074.203, +112.000],
		[+112.000, -093.786, -018.214],
	])
	a = np.array([16, 128, 128])

	return np.dot(image, m.T) + a

def random_triangular_matrix(size, lower=True):
	# type: (int, bool) -> np.ndarray

	""" Returns a triangular matrix with random value between 0 and 1 uniformly.
	"""

	a = np.random.uniform(0, 1, (size, size))
	if lower:
		ind = np.triu_indices(5, 1)
	else:
		ind = np.tril_indices(5, 1)
	a[ind] = 0

	return a

def issquare(A):
	# type: (np.ndarray, ) -> bool

	if len(A.shape) != 2:
		return False

	if A.shape[0] != A.shape[1]:
		return False

	return True

def batch_vTAv(A, v):
	# type: (float[B, X, X], float[B, X]) -> float[B]

	""" Performs batched calculation of `v^T A v` transform.
		Special case of bilinear form `x^T A y`
	"""

	""" Faster than
		Av = np.matmul(A, v[...,:,None]) # [B, X, 1]
		return np.matmul(v[...,None,:], Av).squeeze((-2, -1)) # [B]
	"""

	return np.einsum("...k,...kl,...l->...", v, A, v)

# was: inner1d
def batch_inner(a, b):
	# type: (np.ndarray, np.ndarray) -> np.ndarray

	""" Performs a batched inner product over the last dimension.
		Replacement for deprecated `from numpy.core.umath_tests import inner1d`.
		Shapes: (A, B), (A, B) -> (A, )
	"""

	if a.shape != b.shape:
		raise ValueError("All dimensions have to be equal")

	return np.einsum("...i,...i->...", a, b) # faster than np.sum(a * b, axis=-1)

def batch_outer(a, b):
	# type: (np.ndarray, np.ndarry) -> np.ndarray

	""" Performs a batched outer product over the last dimension.
		Shapes: (A, B), (A, C) -> (A, B, C)
	"""

	if a.shape[:-1] != b.shape[:-1]:
		raise ValueError("All except the last dimension have to be equal")

	return np.einsum("...i,...j->...ij", a, b) # slightly faster than np.multiply(a[...,:,None], b[...,None,:])

def batchtopk(probs, k=None, axis=-1, reverse=False):
	# type: (np.ndarray, np.ndarray, int) -> np.ndarray

	""" `probs` values ndarray
		`k` take the smallest `k` elements, if `reverse` is False
			and the largest `k` if `reverse` is True
		`axis` sorting and selection axis.
	"""

	assert k is None or k > 0, "k must be larger than zero. Use None to chose all elements."
	assert axis == -1, "Only last axis supported atm"
	assert len(probs.shape) > 1

	if reverse:
		sign = -1
	else:
		sign = 1

	indices = np.argsort(sign * probs, axis=-1) # use argpartition?
	probs = np.take_along_axis(probs, indices[...,:k], axis=-1)

	return indices, probs

def logtrace(m):
	""" Calcuates the sum of the logs of the diagonal elements (batchwise if neccessary)
		m: [..., x, x]
	"""

	return np.sum(np.log(np.diagonal(m, axis1=-2, axis2=-1)), axis=-1)

def shiftedexp(pvals):
	# type: (np.ndarray, ) -> np.ndarray

	""" Prevents overflow. Can be used if probabilities are normalized again later.
	"""

	return np.exp(pvals - np.max(pvals))

class Sampler(object):

	def __init__(self, cdf):
		self.cdf = cdf
		self.psum = cdf[-1]

	def __call__(self):
		rand = np.random.uniform(0, self.psum)
		return np.searchsorted(self.cdf, rand, side="right")

def sample_probabilities(pvals):
	# type: (np.ndarray, ) -> Callable[[], int]

	""" Sample from list of probabilities `pvals` with replacement.
		The probabilities don't need to be normalized.
	"""

	return Sampler(np.cumsum(pvals))

class UnboundedSparseMatrix(object):

	def __init__(self, dtype=float):
		# type: (type, ) -> None

		self.dtype = dtype
		self.zero = self.dtype(0)
		self.m = dict()
		self.cols = 0
		self.rows = 0

	def __getitem__(self, slice):
		# type: (tuple, ) -> T

		return self.m.get(slice, self.zero)

	def __setitem__(self, slice, value):
		# type: (tuple, T) -> None

		c, r = slice
		self.cols = max(self.cols, c+1)
		self.rows = max(self.rows, r+1)
		self.m[slice] = value

	def todense(self):
		# type: () -> np.ndarray

		ret = np.zeros((self.cols, self.rows), dtype=self.dtype)

		for slice, value in viewitems(self.m):
			ret[slice] = value

		return ret

def normalize(pvals):
	# type: (np.ndarray, ) -> np.ndarray

	return pvals / np.sum(pvals)

def categorical(pvals):
	# type: (np.ndarray, ) -> int

	""" Requires normalized inputs: sum(pvals) ~= 1 """

	return np.argmax(np.random.multinomial(1, pvals))

def inf_matrix_power(pm):
	w, v = np.linalg.eig(pm) # scipy.linalg.eig would probably by faster as it can return the left and right eigen vectors

	if not np.isclose(w[0], 1.):
		raise ValueError("The first eigenvalue is not none. Is this a right stochastic matrix?")

	vi = np.linalg.inv(v)
	d = np.zeros(m.shape[0], dtype=np.float)
	d[0] = 1.

	return np.matmul(v, np.matmul(np.diag(d), vi))

def decompress(selectors, data, default):
	# type: (np.ndarray[bool], np.ndarray[T], T) -> None

	""" Same result as:
		`np.array(list(genutility.iter.decompress(selectors, iter(data), default)))`
		but faster
	"""

	out = np.full(len(selectors), default)
	out[selectors] = data
	return out

def unblock(arr, n1, n2, axis1=-1, axis2=-2, blocksize=False):

	""" Inverse of np.block.
		Set axis to (-2, -1) to modify the order of the result.
	"""

	""" test (stackoverflow): Ok, so considering I have N block matrices with bm x bn dimension and want to stack them in a m x n matrix, provided N = m x n, I would then have x.reshape(m,n,bm,bn).swapaxes(1,2).reshape(bm*m,-1)
	"""

	s = np.array(arr.shape)
	assert s[axis1] % n1 == 0 and s[axis2] % n2 == 0, "{}x{} does not divide by {}x{}".format(s[axis1], s[axis2], n1, n2)

	if blocksize:
		n1 = s[axis1] // n1
		n2 = s[axis2] // n2

	arr = np.array(np.split(arr, n1, axis1))
	arr = np.array(np.split(arr, n2, axis2))

	inv_blocksize = n1 * n2
	total = s[axis1] * s[axis2]
	s[axis2] = inv_blocksize
	s[axis1] = total // inv_blocksize

	return np.reshape(arr, s)

def remove_color(img, ratio, neutral_color=RGB_WHITE):
	# type: (np.ndarray, float, Tuple[int, int, int]) -> np.ndarray

	""" Replace colored pixels with a `neutral_color`. The `ratio` defines the 'colorfulness' above
		which level the pixel should be replace.
		I.e. if the `ratio` is 1 nothing will be replaced,
		if `ratio` is 0 only strict greys are kept unmodified.
	"""

	channels = img.shape[-1]
	assert channels == 3, "Not a 3 channel color image"

	norm = np.std(np.array(RGB_YELLOW)) # this is the same for all pure colors

	sd = np.std(img, axis=-1)
	img[sd > ratio*norm] = neutral_color

#@opjit(cache=False) doesn't help a lot, because it's a generator
def sliding_window_2d(image, window_size, step_size=(1, 1)):
	# type: (np.ndarray, Tuple[int, int], Tuple[int, int]) -> Iterator[np.ndarray]

	win_x, win_y = window_size
	step_x, step_y = step_size
	height, width = image.shape[0:2]

	for y in range(0, height - win_y + 1, step_y):
		for x in range(0, width - win_x + 1, step_x):
			yield image[y:y + win_y, x:x + win_x, ...]

def histogram_correlation(hist1, hist2):
	# type: (np.ndarray, np.ndarray) -> np.ndarray

	""" Input shape of `hist1` and `hist2`: [batch..., levels]. The correlation is calculated over `levels`
		and then batched over the remaining dimensions.
	"""

	assert len(hist1.shape) >= 1 and hist1.shape == hist2.shape

	h1norm = (hist1.T - np.mean(hist1, axis=-1)).T
	h2norm = (hist2.T - np.mean(hist2, axis=-1)).T

	num = np.sum(h1norm * h2norm, axis=-1)
	denom = np.sqrt(np.sum(h1norm**2, axis=-1)*np.sum(h2norm**2, axis=-1))
	return num / denom

if __name__ == "__main__":
	import timeit
	from .numpy import random_triangular_matrix

	image = np.random.randint(0, 255, (1000, 1000))

	list(sliding_window_2d(image, (100, 100), (1, 1))) # warmup
	print(min(timeit.repeat('list(sliding_window_2d(image, (100, 100), (1, 1)))', number=10, repeat=5, globals=globals())))



from __future__ import absolute_import, division, print_function, unicode_literals

import socket
import netifaces

def get_standard_gateway(default=None):
	for ipv4_gateway in netifaces.gateways()["default"][netifaces.AF_INET]:
		try:
			socket.inet_aton(ipv4_gateway)
			return ipv4_gateway
		except socket.error:
			pass
	return default



from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

def hamming_distance(a, b):
	# type: (bytes, bytes) -> int

	a = np.unpackbits(np.frombuffer(a, dtype=np.uint8))
	b = np.unpackbits(np.frombuffer(b, dtype=np.uint8))
	return np.count_nonzero(a != b)



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import zip

from functools import reduce
from operator import mul
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Iterable, TypeVar
	from ..typing import Computable
	T = TypeVar("T")

try:
	from math import nan
except ImportError:
	nan = float("NaN")

try:
	from math import prod

except ImportError:

	def prod(it, start=1): # type: ignore
		# type: (Iterable[T], T) -> T

		""" Counterpart to built-in `sum()`. """

		return reduce(mul, it, start)

	def prod_2(it, start=1):
		# type: (Iterable[Computable], Computable) -> Computable

		""" Counterpart to built-in `sum()`. """

		for x in it:
			start *= x
		return start

try:
	from math import dist
except ImportError:

	from math import sqrt

	def dist(p, q):
		return sqrt(sum((px - qx) ** 2. for px, qx in zip(p, q)))

try:
	from math import isqrt

except ImportError:

	try:
		import gmpy2

		def isqrt(n):
			""" Integer square root of `n`. """

			return int(gmpy2.isqrt(gmpy2.mpz(n)))

	except ImportError:
		def isqrt(n):
			""" Integer square root of `n`.
				It uses float conversion and is not accurate for large numbers.
				Install `gmpy2` for improved speed and accuracy.
			"""

			return int(n**0.5)



from __future__ import absolute_import, division, print_function, unicode_literals

import os.path

from .twothree import FileNotFoundError
from .compat.importlib.util import find_spec

from .toml import read_toml
from .json import read_json_schema
from .os import get_appdata_dir

if __debug__:
	import jsonschema

def _load(name):
	# type: (str, ) -> Dict[str, Any]

	configfilename = name+".toml"

	# try appdata directory
	try:
		return read_toml(os.path.join(get_appdata_dir(), name, configfilename))
	except FileNotFoundError:
		pass

	# try module directory
	try:
		spec = find_spec(name)

		if spec is None:
			raise ImportError("No module named '{}'".format(name)) # or FileNotFoundError?

		if spec.has_location:
			modpath = os.path.dirname(spec.origin)
		else:
			try:
				modpath = spec.submodule_search_locations[0]
			except (TypeError, IndexError):
				raise FileNotFoundError

		return read_toml(os.path.join(modpath, configfilename))
	except (ImportError, FileNotFoundError):
		pass

	# try working directory
	try:
		return read_toml(configfilename)
	except FileNotFoundError:
		raise FileNotFoundError("{} could not be found in application data, module path or current directory".format(configfilename))

def load(name, json_schema=None):
	# type: (str, Optional[Union[dict, str]]) -> Dict[str, Any]

	obj = _load(name)

	if json_schema is None:
		return obj

	from jsonschema import validate

	if isinstance(json_schema, str):
		json_schema = read_json_schema(json_schema)

	validate(obj, json_schema)
	return obj



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from datetime import datetime

import win32evtlog, winerror
import wmi

from .math import NaN

logger = logging.getLogger(__name__)

class Resource(object):

	def __init__(self, wmii=None):
		if wmii:
			self.wmii = wmii
		else:
			self.wmii = wmi.WMI()

	def __iter__(self):
		return self

class os_resources(Resource):

	def __next__(self):
		threads = 0
		handles = 0
		pagefaults = 0
		for proc in self.wmii.Win32_Process():
			threads += proc.ThreadCount
			handles += proc.HandleCount
			pagefaults += proc.PageFaults
		return threads, handles, pagefaults

class cpu_usage(Resource):

	def __init__(self, wmii=None, proc_name="_Total"):

		""" proc_name is the name of the processor,
			default value mean across all processors.
		"""

		Resource.__init__(self, wmii)
		self.proc_name = proc_name

		self.real0 = NaN
		self.user0 = NaN
		self.sys0 = NaN
		self.ts0 = NaN

	@staticmethod
	def get_proc_names(wmii):
		return [proc.name for proc in wmii.Win32_PerfRawData_PerfOS_Processor()]

	def __next__(self):
		p, = self.wmii.Win32_PerfRawData_PerfOS_Processor(name=self.proc_name)

		real = int(p.PercentProcessorTime)
		user = int(p.PercentUserTime)
		sys = int(p.PercentPrivilegedTime)
		ts = int(p.Timestamp_Sys100NS)

		try:
			delta_time = ts - self.ts0
			real_ratio = 1 - (real - self.real0) / delta_time # One Minus should not be necessary
			user_ratio = (user - self.user0) / delta_time
			sys_ratio = (sys - self.sys0) / delta_time
		except ZeroDivisionError:
			real_ratio = NaN
			user_ratio = NaN
			sys_ratio = NaN

		self.real0 = real
		self.user0 = user
		self.sys0 = sys
		self.ts0 = ts

		return (real_ratio, user_ratio, sys_ratio, delta_time/10000000)

class ram_usage(Resource):

	def __next__(self):
		p, = self.wmii.Win32_OperatingSystem()
		total = int(p.TotalVisibleMemorySize)
		free = int(p.FreePhysicalMemory)
		assert total > 0
		return 1 - (free / total)

class hdd_usage(Resource):

	def __next__(self):
		ret = []
		for disk in self.wmii.Win32_LogicalDisk(DriveType=3):
			total = int(disk.Size)
			free = int(disk.FreeSpace)
			ret.append((disk.Caption, 1 - (free / total)))
		return ret

def event_logs(server="localhost", source="System"):
	# type: (str, str) -> Iterator[dict]

	"""
	EventType: severity level

	EVENTLOG_ERROR_TYPE 0x0001
	EVENTLOG_WARNING_TYPE 0x0002
	EVENTLOG_INFORMATION_TYPE 0x0004
	EVENTLOG_AUDIT_SUCCESS 0x0008
	EVENTLOG_AUDIT_FAILURE 0x0010

	EventID: same for same class of messages, stringinserts are provides to template belonging to eventid

	"""

	fields = {"Reserved", "RecordNumber", "TimeGenerated", "TimeWritten", "EventID", "EventType", "EventCategory", "ReservedFlags", "ClosingRecordNumber", "SourceName", "StringInserts", "Sid", "Data", "ComputerName"}
	fields = {"Reserved", "RecordNumber", "TimeGenerated", "TimeWritten", "EventID", "EventType", "EventCategory", "ReservedFlags", "ClosingRecordNumber", "SourceName", "StringInserts", "ComputerName"}

	# missing: "Sid", "Data"

	handle = win32evtlog.OpenEventLog(server, source)
	flags = win32evtlog.EVENTLOG_SEQUENTIAL_READ | win32evtlog.EVENTLOG_FORWARDS_READ

	logger.debug("Preparing to read {} events".format(win32evtlog.GetNumberOfEventLogRecords(handle)))

	try:
		while True:
			events = win32evtlog.ReadEventLog(handle, flags, 0)
			if events:
				for event in events:
					ret = {f:getattr(event, f) for f in fields}
					# does that work for different time zones?
					ret["TimeGenerated"] = datetime.fromtimestamp(event.TimeGenerated.timestamp())
					ret["TimeWritten"] = datetime.fromtimestamp(event.TimeWritten.timestamp())
					if ret["StringInserts"]:
						ret["StringInserts"] = ", ".join(ret["StringInserts"]).replace("\r\n", "\n")
					if not ret["StringInserts"]:
						ret["StringInserts"] = None
					ret["EventID"] = winerror.HRESULT_CODE(ret["EventID"])
					#print(ret)
					yield ret
			else:
				break
	finally:
		win32evtlog.CloseEventLog(handle)

if __name__ == "__main__":
	from builtins import zip
	from time import sleep
	from argparse import ArgumentParser

	from genutility.stdio import print_line

	parser = ArgumentParser(description='Monitor OS resource usage.')
	args = parser.parse_args()

	wmii = wmi.WMI()

	for a, b, c, d in zip(cpu_usage(wmii), ram_usage(wmii), hdd_usage(wmii), os_resources(wmii)):
		print("CPU:", a)
		print("RAM:", b)
		print("HDD:", c)
		print("OS:", d)
		print_line()
		sleep(1)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import input, str
from future.utils import PY3

import sys, logging
from time import sleep

try:
	from shutil import get_terminal_size
	_terminal_width = get_terminal_size((80, 30)).columns
except ImportError: # < Python 3.3
	_terminal_width = 80

def print_line(char="-", length=_terminal_width, file=None, flush=False):
	# type: (str, int, IO, bool) -> None
	""" file=None, defaults to sys.stdout in print(). print() handles sys.stdout==None fine. """

	print(char*length, end="", file=file, flush=flush)

def safe_input(s, block=10):
	""" makes sure ctrl-c on input() only raises KeyboardInterrupt and not EOFError+KeyboardInterrupt.
		Waits at most `block` seconds to be sure.
	"""

	try:
		return input(s)
	except EOFError:
		sleep(block) # can be interrupted by KeyboardInterrupt

def info_print(msg=None, args=tuple(), exception=None):
	# not (msg or exception) this doesn't do anything

	if exception and msg:
		logging.exception(msg % args, exc_info=exception)
		#logging.exception(msg, *args, exc_info=exception) this fails for some weird reason
	else:
		if exception:
			logging.exception("Unhandled exception", exc_info=exception)
		if msg:
			print(msg % args)

if PY3:
	def input_ex(s, file_out=sys.stdout, file_in=sys.stdin):
		# type: (str, filelike, filelike) -> str

		print(s, file=file_out, end="")
		file_out.flush()
		return sys.stdin.readline().rstrip("\n")

else:
	def input_ex(s, file_out=sys.stdout, file_in=sys.stdin):
		# type: (str, filelike, filelike) -> str

		print(s.encode(file_out.encoding, errors="replace"), file=file_out, end=b"")
		return sys.stdin.readline().decode(file_in.encoding).rstrip("\n")

# was: raw_input_ex
def input_type(s, type=None, predicate=None, errormsg=None, exception=Exception, file=sys.stdout):
	while True:
		ret = input_ex(s, file_out=file)
		if predicate and not predicate(ret):
			if errormsg:
				print(errormsg % (ret,))
			continue
		if type:
			try:
				return type(ret)
			except exception:
				if errormsg:
					print(errormsg % (ret,))
				continue
		return ret

def confirm(msg, yesno=True):
	if yesno:
		s = input_type(msg+" (yes, no): ", predicate=lambda x: x.strip().lower() in {"yes", "no"})
		return s.strip().lower() == "yes"
	else:
		return input_type(msg+" (anything, nothing): ", type=bool) # bool cannot throw if given string I think...

def waitquit(msg=None, args=tuple(), exception=None): # fixme: ferutility.printing breaks this
	info_print(msg, args, exception)
	input("Press enter to exit...")
	sys.exit(msg or exception)

def waitcontinue(msg=None, args=tuple(), exception=None):
	info_print(msg, args, exception)
	input("Press enter to continue...")

def errorquit(msg=None, args=tuple(), exception=None):
	info_print(msg, args, exception)
	sys.exit(msg or exception)



from __future__ import absolute_import, division, print_function, unicode_literals

import bz2, warnings
from io import TextIOWrapper

"""
most of his file is copied and modified from Python35/Lib/bz2.py file
"""

"""
import threading
tl = threading.local()
tl.indent = 0
"""

from future.utils import PY2

if PY2:
	warnings.warn("Python2 bz2 module does not support multiple streams", ImportWarning, stacklevel=2)

	class BZ2File3(bz2.BZ2File):

		"""
		Changed in version 3.1: Support for the with statement was added.
		Changed in version 3.3: The fileno(), readable(), seekable(), writable(), read1() and readinto() methods were added.
		"""

		def __init__(self, filename, mode="r", buffering=0, compresslevel=9):
			bz2.BZ2File.__init__(self, filename, mode, buffering, compresslevel)
		
		"""
		def __getattribute__(self, name):
			import logging
			attr = object.__getattribute__(self, name)
			if callable(attr):
				def wrapper(*args, **kwargs):
					try:
						indent = tl.indent
						tl.indent += 1
						ret = attr(*args, **kwargs)
						tl.indent -= 1
						logging.warning("%s%s(%s, %s): %s", "--"*indent, name, args, kwargs, ret)
						return ret
					except Exception as e:
						logging.exception(name)
						raise
				return wrapper
			else:
				return attr
		"""

		def fileno(self):
			assert False

		def seekable(self):
			return self.readable()

		def readable(self):
			assert not self.closed
			return "r" in self.mode

		def writable(self):
			assert not self.closed
			return "w" in self.mode

		def flush(self): # needed by fileinput.py
			pass

		def read1(self, size=-1):
			return self.read(size)

		def readinto(self, b):
			assert False

		def __enter__(self):
			return self

		def __exit__(self, exc_type, exc_value, traceback):
			self.close()

	def open(filename, mode="rb", compresslevel=9, encoding=None, errors=None, newline=None):

		if "t" in mode:
			if "b" in mode:
				raise ValueError("Invalid mode: %r" % (mode,))
		else:
			if encoding is not None:
				raise ValueError("Argument 'encoding' not supported in binary mode")
			if errors is not None:
				raise ValueError("Argument 'errors' not supported in binary mode")
			if newline is not None:
				raise ValueError("Argument 'newline' not supported in binary mode")

		bz_mode = mode.replace("t", "")
		binary_file = BZ2File3(filename, bz_mode, compresslevel=compresslevel)

		if "t" in mode:
			return TextIOWrapper(binary_file, encoding, errors, newline)
		else:
			return binary_file

else:
	open = bz2.open



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

try:
	from contextlib import nullcontext # pylint: disable=no-name-in-module

except ImportError:

	if TYPE_CHECKING:
		from typing import TypeVar, Optional
		T = TypeVar("T")

	try:
		from contextlib import AbstractContextManager
	except ImportError:
		AbstractContextManager = object

	class nullcontext(AbstractContextManager):

		""" A context manager which doesn't do anything except return its argument.
			Kind of like an identity function for contexts, useful as default argument.
		"""

		def __init__(self, enter_result=None):
			# type: (Optional[T], ) -> None

			self.enter_result = enter_result

		def __enter__(self):
			# type: () -> Optional[T]

			return self.enter_result

		def __exit__(self, exc_type, exc_value, traceback):
			pass



from __future__ import absolute_import, division, print_function, unicode_literals

import os.path
from sys import version_info
from io import open, TextIOWrapper, TextIOBase, RawIOBase, BufferedIOBase, SEEK_SET, SEEK_END
from typing import TYPE_CHECKING

from .iter import iter_equal, consume, resizer
from .math import PosInfInt
from .ops import logical_xor, logical_implication

if __debug__:
	from .compat import gzip, bz2

if TYPE_CHECKING:
	from typing import Callable, Optional, Union, IO, TextIO, BinaryIO

FILE_IO_BUFFER_SIZE = 8*1024*1024

def read_file(path, mode="b", encoding=None, errors=None):
	# type: (str, str, Optional[str]) -> bytes

	""" Reads and returns whole file. If content is not needed use consume_file()"""

	assert (encoding is None) == ("b" in mode)

	if not "r" in mode:
		mode = "r" + mode

	with open(path, mode, encoding=encoding, errors=errors) as fr:
		return fr.read()

def write_file(data, path, mode="wb", encoding=None, errors=None):
	# type: (str, Union[str, bytes], str, Optional[str]) -> None

	""" Writes file. """

	assert (encoding is None) == ("b" in mode)

	with open(path, mode, encoding=encoding, errors=errors) as fw:
		fw.write(data)

def linefileiter(fin, check_last=True): # kinda unnecessary
	last = None
	for line in fin:
		yield line
		last = line
	assert last.endswith("\n"), "File does not end in newline: {}".format(last)

def get_file_range(path, start, size):
	# type: (str, int, int) -> bytes

	with open(path, "rb") as fp:
		fp.seek(start)
		return fp.read(size)

def truncate_file(path, size):
	# type: (str, int) -> None

	with open(path, "r+b") as fp:
		fp.truncate(size)

def wrap_text(bf, mode, encoding, errors, newline):
	if "t" in mode:
		return TextIOWrapper(bf, encoding, errors, newline)
	return bf

def copen(file, mode="rt", archive_file=None, encoding=None, errors=None, newline=None, compresslevel=9):
	# type: (str, str, Optional[str], Optional[str], Optional[str], Optional[str], int) -> IO

	"""
	`compresslevel`: 0-9, 0: no compression, 1: least, 9: highest compression
	"""

	is_text = "t" in mode
	is_binary = "b" in mode

	assert logical_xor(is_text, is_binary), "Explicit text or binary mode required: {}".format(mode)
	assert logical_implication(is_binary, encoding is None and errors is None)

	if is_text:
		encoding = encoding or "utf-8"
		errors = errors or "strict"

	ext = os.path.splitext(file)[1].lower()
	if ext == '.gz':
		from .compat import gzip
		return gzip.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

	elif ext == '.bz2':
		from .compat import bz2
		return bz2.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

	elif ext == '.zip':
		assert archive_file, "archive_file must be specified"
		from zipfile import ZipFile

		newmode = "".join(set(mode) - {"t", "b"})

		with ZipFile(file, newmode) as zf: # note: even if the outer zip file is closed, the inner file can still be read apparently
			if version_info >= (3, 6):
				bf = zf.open(archive_file, newmode, force_zip64=True)
			else:
				bf = zf.open(archive_file, newmode)

		return wrap_text(bf, mode, encoding, errors, newline)

	return open(file, mode, encoding=encoding, errors=errors, newline=newline)

# was: OpenAndDeleteOnError, OpenFileRemoveOnException
class OpenFileAndDeleteOnError(object):

	""" Context manager which opens a file using the same arguments as `open`,
		but deletes the file in case an exception occurs after opening.
	"""

	def __init__(self, file, mode="rt", encoding=None, errors=None, newline=None, compresslevel=9):
	# type: (str, str, Optional[str], Optional[str], Optional[str], int) -> None

		is_text = "t" in mode
		is_binary = "b" in mode

		assert logical_xor(is_text, is_binary), "Explicit text or binary mode required: {}".format(mode)
		assert logical_implication(is_binary, encoding is None and errors is None)

		if is_text:
			encoding = encoding or "utf-8"
			errors = errors or "strict"

		self.file = file
		self.mode = mode
		self.encoding = encoding
		self.errors = errors
		self.newline = newline
		self.compresslevel = compresslevel
		self.fp = None # type: Optional[IO]

	def __enter__(self):
		# type: () -> IO

		self.fp = copen(self.file, self.mode, None, self.encoding, self.errors, self.newline, self.compresslevel)
		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		self.fp.close()
		if exc_type:
			# fixme: race condition
			# use https://stackoverflow.com/a/3594593 and `ReOpenFile` on Windows
			# or SetFileInformationByHandle
			# see also: https://nullprogram.com/blog/2016/08/07/
			os.remove(self.file)

class PathOrBinaryIO(object):

	def __init__(self, fname, mode="rb", close=False):
		# type: (Union[str, BinaryIO],  str, str, str, Optional[str], bool) -> None

		if isinstance(fname, (RawIOBase, BufferedIOBase)):
			self._close = close
			self.fp = fname
		else:
			self._close = True
			self.fp = copen(fname, mode)

	def __enter__(self):
		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		if self._close:
			self.fp.close()

class PathOrTextIO(object):

	def __init__(self, fname, mode="rt", encoding="utf-8", errors="strict", newline=None, close=False):
		# type: (Union[str, TextIO],  str, str, str, Optional[str], bool) -> None

		if isinstance(fname, TextIOBase):
			self.close = close
			self.fp = fname
		else:
			self.close = True
			self.fp = copen(fname, mode, encoding=encoding, errors=errors, newline=newline)

	def __enter__(self):
		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		if self.close:
			self.fp.close()

class LastLineFile(object):

	chunk_size = 1024 * 4
	nl = "\n"

	def __init__(self, path, mode="rt+"):
		self.f = open(path, mode)
		self.ll_pos = None

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()

	def close(self):
		self.f.close()

	def _seek_to_last_line(self):
		if self.ll_pos is None:
			self.ll_pos = self._get_last_line_pos()

		self.f.seek(self.ll_pos, SEEK_SET)

	def _get_last_line_pos(self):
		self.f.seek(0, SEEK_END)
		pos = self.f.tell()
		ret_pos = 0
		while pos > 0:
			seek_next = max(pos - self.chunk_size, 0)
			self.f.seek(seek_next, SEEK_SET)
			chunk = self.f.read(pos - seek_next)
			fpos = chunk.rfind("\n")
			if fpos != -1:
				ret_pos = seek_next + fpos + 1
				if ret_pos != pos:
					break
			pos = seek_next
		return ret_pos

	def _validate(self, s):
		if self.nl in s:
			raise ValueError("Newline in input")

	def read(self):
		self._seek_to_last_line()
		return self.f.read()

	def newline(self, s):
		self.f.seek(0, SEEK_END)
		ret = self.f.write(self.nl)
		self.ll_pos = self.f.tell()
		return ret

	def replace(self, s):
		self._validate(s)
		self._seek_to_last_line()
		ret = self.f.write(s)
		self.f.truncate()
		return ret

class Tell(object):

	def __init__(self, fp):
		# type: (FileLike, ) -> None

		try:
			assert fp.seekable() == False
		except AttributeError:
			pass

		self._fp = fp
		self._pos = 0

	# wrapped

	def read(self, size=None): # the docs for `io.BufferedIOBase` say that size=-1, but that's not true for `HTTPResponse`
		ret = self._fp.read(size)
		self._pos += len(ret)
		return ret

	def read1(self, n=-1):
		ret = self._fp.read1(n)
		self._pos += len(ret)
		return ret

	def readall(self):
		ret = self._fp.readall()
		self._pos += len(ret)
		return ret

	def readline(self, limit=-1):
		ret = self._fp.readline(limit)
		self._pos += len(ret)
		return ret

	def readlines(self, hint=-1):
		raise NotImplementedError
		return self._fp.readlines(hint)

	def readinto(self, b):
		ret = self._fp.readinto(b)
		self._pos += ret
		return ret

	def readinto1(self, b):
		ret = self._fp.readinto1(b)
		self._pos += ret
		return ret

	def write(self, b): # fixme: should this be added to _pos?
		return self._fp.write(b)

	def writelines(self, lines): # fixme: should this be added to _pos?
		return self._fp.writelines(lines)

	def seek(self, offset, whence=SEEK_SET):
		raise IOError("Stream is not seekable")

	def tell(self):
		return self._pos

	# redirected

	def close(self):
		return self._fp.close()

	def detach(self):
		return self._fp.detach()

	def fileno(self):
		return self._fp.fileno()

	def flush(self):
		return self._fp.flush()

	def isatty(self):
		return self._fp.isatty()

	def readable(self):
		return self._fp.readable()

	def seekable(self):
		return self._fp.seekable()

	def truncate(self, size=None):
		return self._fp.truncate(size)

	def writable(self):
		return self._fp.writable()

from future.moves.urllib import response
from collections import deque

class BufferedTell(response.addinfourl): # fixme: untested!!!

	def __init__(self, filesize):
		#response.addinfourl.__init__(self, *args, **kwargs)
		self._pos = 0
		self._read = 0
		self.filesize = filesize
		self.buf = deque(maxlen=1024)

	def read(self, num=-1):
		if num == -1:
			num = self.filesize - self._pos
			print("num=-1")
		if self._pos == self._read:
			self._pos += num
			self._read += num
			data = response.addinfourl.read(num)
			self.buf.extend(data)
			return data
		elif self._pos < self._read:
			delta = self._read - self._pos
			self._pos += num
			if num <= delta:
				data = b"".join(self.buf.pop() for i in range(delta))
				self.buf.extend(data[::-1])
				return data[-1:-1-num:-1]
			else:
				datab = b"".join(self.buf.pop() for i in range(delta))
				datab = datab[::-1]
				self.buf.extend(datab)
				dataf = response.addinfourl.read(num - delta)
				self.buf.extend(dataf)
				return datab + dataf
		else:
			#return b""
			raise Exception("Unbuffered read")

	def tell(self):
		return self._pos

	def seek(self, offset, whence=0):
		if whence == 0:
			self._pos = offset
		elif whence == 1:
			self._pos += offset
		elif whence == 2:
			self._pos = self.filesize - offset
		else:
			raise ValueError("Unknown whence")

def copyfilelike(fin, fout, amount=None, buffer=FILE_IO_BUFFER_SIZE, report=None):
	# type: (IO, IO, Optional[int], Optional[int], Optional[Callable]) -> int

	""" Read data from `fin` in chunks of size `buffer` and write them to `fout`.
		Optionally limit the amout of data to `amount`. `report` can be a callable which receives
		the total number of bytes copied and bytes remaining.
		see `shutil.copyfileobj`
	"""

	# todo: have different input and output buffers

	amount = amount or PosInfInt

	copied = 0
	while amount > 0:
		if report:
			report(copied, amount+copied)

		data = fin.read(min(buffer, amount))

		if not data:
			break
		fout.write(data)
		amount -= len(data)
		copied += len(data)
	return copied

def simple_file_iter(fr, chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (IO, int) -> Iterator

	""" Iterate file-like object `fr` and yield chunks of size `chunk_size`. """

	assert isinstance(chunk_size, int), "chunk_size needs to be an integer"

	#return iter(partial(fr.read, chunk_size), "") sentinel cannot be defined well
	while True:
		data = fr.read(chunk_size)
		if data:
			yield data
		else:
			break

def reversed_file_iter(fp, chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (IO, int) -> Iterator

	""" Generate blocks of file's contents in reverse order. """

	fp.seek(0, SEEK_END)
	here = fp.tell()
	while here > 0:
		delta = min(chunk_size, here)
		here -= delta
		fp.seek(here, SEEK_SET)
		yield fp.read(delta)

def limited_file_iter(fr, amount, chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (IO, int, int) -> Iterator

	""" Iterate file-like object `fr` and yield chunks of size `chunk_size`.
		Limit output to `amount` bytes.
	"""

	while amount > 0:
		data = fr.read(min(chunk_size, amount))
		if not data:
			break
		yield data
		amount -= len(data)

def iterfilelike(fr, amount=None, chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (IO, Optional[int], int) -> Iterator

	""" Iterate file-like object `fr` and yield chunks of size `chunk_size`.
		Optionally limit output to `amount` bytes.
	"""

	if amount is None:
		return simple_file_iter(fr, chunk_size)
	else:
		return limited_file_iter(fr, amount, chunk_size)

def blockfileiter(path, mode="rb", encoding=None, errors=None, amount=None, chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (Path, str, Optional[str], Optional[int], int) -> Iterator

	""" Iterate over file at `path` and yield chunks of size `chunk_size`.
		Optionally limit output to `amount` bytes.
	"""

	assert (mode == "rt" and encoding) or (mode == "rb" and not encoding and not errors)
	with open(path, mode, encoding=encoding, errors=errors) as fr:
		for data in iterfilelike(fr, amount, chunk_size):
			yield data

def bufferedfileiter(path, chunk_size, mode="rb", encoding=None, errors=None, amount=None, buffer_size=FILE_IO_BUFFER_SIZE):
	# type: (Path, int, str, Optional[str], Optional[int], int) -> Iterator

	""" Iterate over file at `path` reading chunks of size `buffer_size` at a time.
		Yields data chunks of `chunk_size` and optionally limits output to `amount` bytes.
	"""

	return resizer(blockfileiter(path, mode, encoding, errors, amount, buffer_size), chunk_size)

def byte_out(path, buffer_size=FILE_IO_BUFFER_SIZE):
	# type: (str, int) -> Iterator[bytes]
	""" open file and return it byte by byte """

	return resizer(blockfileiter(path, mode="rb", chunk_size=buffer_size), 1)

def consume_file(filename, buffer_size=FILE_IO_BUFFER_SIZE):
	# type: (str, int) -> None
	""" reads whole file but ignores content """

	consume(blockfileiter(filename, mode="rb", chunk_size=buffer_size))

# was: same_files, textfile_equal: equal_files(*paths, mode="rt")
def equal_files(*paths, **kwargs):
	# type: (*str, str, int) -> bool

	""" Check if files at `*paths` are equal. Chunks of size `chunk_size` are read at a time.
		Data can be optionally limited to `amount`.
	"""

	# python2 fix for 'equal_files(*paths, mode="rb", encoding=None, errors=None, amount=None, chunk_size=FILE_IO_BUFFER_SIZE)'
	mode = kwargs.pop("mode", "rb")
	encoding = kwargs.pop("encoding", None)
	errors = kwargs.pop("errors", None)
	amount = kwargs.pop("amount", None)
	chunk_size = kwargs.pop("chunk_size", FILE_IO_BUFFER_SIZE)
	assert not kwargs, "Invalid keyword arguments"

	its = tuple(blockfileiter(path, mode=mode, encoding=encoding, errors=errors, amount=amount, chunk_size=chunk_size) for path in paths)
	return iter_equal(*its)

def is_all_byte(fr, thebyte=b"\0", chunk_size=FILE_IO_BUFFER_SIZE):
	# type: (IO, bytes, int) -> bool

	""" Test if file-like `fr` consists only of `thebyte` bytes. """

	assert isinstance(thebyte, bytes)

	thebyte = thebyte*chunk_size
	for data in simple_file_iter(fr, chunk_size):
		if data != thebyte[:len(data)]:
			return False
	return True

# is this still needed?
def file_byte_reader(filename, inputblocksize, outputblocksize, DEBUG=True):
	assert (inputblocksize % outputblocksize == 0) or (outputblocksize % inputblocksize == 0), "Neither input nor output size is a multiple of the other"

	bytes_yielded = 0
	bytes = bytearray(max(inputblocksize, outputblocksize))
	bytes_used = 0
	for read in blockfileiter(filename, chunk_size=inputblocksize):
		bytes[bytes_used:bytes_used+len(read)] = read
		#print("read {} bytes to pos [{}:{}]".format(len(read),bytes_used,bytes_used+len(read)))
		bytes_used += len(read)
		pos = 0
		while bytes_used >= outputblocksize:
			yield bytes[pos:pos+outputblocksize]
			bytes_yielded += outputblocksize
			#print("yielded {} bytes from pos [{}:{}]".format(outputblocksize,pos,pos+outputblocksize))
			pos += outputblocksize
			bytes_used -= outputblocksize
	if DEBUG:
		import os.path
		filesize = os.path.getsize(filename)
		if filesize != bytes_yielded:
			print("{} bytes yielded, filesize: {}".format(bytes_yielded, filesize))



from __future__ import absolute_import, division, print_function, unicode_literals

from future.moves.itertools import zip_longest
from builtins import zip, map, range, reversed

import sys, logging
from collections import deque
from itertools import groupby, chain, tee, islice, count, combinations, product
from operator import add
from time import time
from types import GeneratorType
from typing import TYPE_CHECKING

from .exceptions import IteratorExhausted

if TYPE_CHECKING:
	from typing import Any, Callable, Iterable, Iterator, TypeVar, Tuple, Sequence, Optional
	from logging import Logger
	T = TypeVar("T")

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

def progress(it, length=None, refresh=1, file=sys.stdout):
	# type: (Union[Iterable, Collection], Optional[int], float, Optional[TextIO]) -> Iterator

	""" Wraps an iterable `it` to periodically print the progress every `refresh` seconds. """

	""" todo: `from operator import length_hint`
		Use `length_hint(it)` for a guess which might be better than nothing.
		Can return 0.
	"""

	if length is None:
		try:
			length = len(it) # type: ignore # TypeError caught below
		except TypeError:
			pass

	if length is not None:
		lstr = "/{}".format(length)
	else:
		lstr = ""

	last = start = time()
	i = 0
	for i, elm in enumerate(it):
		yield elm
		current = time()
		if current - last > refresh:
			last = current
			duration = current-start
			print("{}{}, running for {} seconds ({:0.2e}/s).".format(i, lstr, int(duration), i/duration), end="\r", file=file)
	print("Finished {} in {} seconds.".format(i, int(last-start)), end="\r", file=file)

class Progress(object):

	""" Cannot create a decorator to do this to generic generators,
		because attributes cannot be set on generator objects.
	"""

	def __init__(self, obj):
		self.obj = obj

	def __iter__(self):
		return progress(self.obj)

	def __len__(self):
		# type: () -> int

		""" Can raise a TypeError if underlying object does not support `len()`.
		"""

		return len(self.obj)

def count_distinct(it):
	# type: (Iterable[Any], ) -> int

	return len(set(it))

def reversedzip(*its):
	return zip(*map(reversed, its))

def flatten(it):
	# type: (Union[tuple, list], ) -> Iterator

	for elm in it:
		if isinstance(elm, (tuple, list)):
			for i in flatten(elm):
				yield i
		else:
			yield elm

def diffsgn(a, b):
	# type: (Number, Number) -> int

	""" Sign of the finite derivative.
		Equal to `-cmp(a, b)` or `cmp(b, a).
	"""

	return (a < b) - (a > b)

def asc_peaks(iterable):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields ascending peaks from `iterable`. """

	it = iter(iterable)
	try:
		cur = next(it)
	except StopIteration:
		return

	for x in it:
		if x < cur:
			yield cur
		cur = x
	yield cur

def extrema(iterable, first=None, last=None, derivatives={1, -1}):
	it = iter(iterable)
	old = next(it)
	old_d = diffsgn(old, next(it))

	if old_d == first:
		yield old

	for new in it:
		new_d = diffsgn(old, new)
		if new_d != old_d and old_d in derivatives:
			yield old
		old = new
		old_d = new_d

	if old_d == last:
		yield old

def peaks(it):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields peaks of `it`. """

	return extrema(it, -1, 1, {1})

def valleys(it):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields valleys of `it`. """

	return extrema(it, 1, -1, {-1})

def empty():
	# type: () -> Iterator

	""" An empty iterator. Other methods include:
		iter(())
		yield from ()
	"""

	return
	yield # pylint: disable=unreachable

def batch(it, n, filter=None):
	# type: (Iterable[Any], int, Optional[Callable]) -> Iterator[Any] # return type cannot be more specific because of filter()

	""" Batches iterator `it` into batches of size `n`.
		Optionally post-processes batch with `filter`.
	"""

	it = iter(it)
	while True:
		chunk_it = islice(it, n)
		try:
			first_el = next(chunk_it)
		except StopIteration:
			return
		if filter:
			yield filter(chain((first_el,), chunk_it))
		else:
			yield chain((first_el,), chunk_it)

def advance(it, n):
	# type: (Iterable, int) -> None
	""" Advances the iterable `it` n steps. """

	"""
	for i in range(n):
		next(it, None)
	"""
	next(islice(it, n, n), None)

def filternone(it):
	# type: (Iterable, ) -> Iterator

	""" Removes all `None` values from `it`. """

	return (i for i in it if i is not None)

def all_equal(it):
	# type: (Iterable, ) -> bool

	""" Returns `True` if all elements of `it` are equal to each other.
		Much faster than:
			return len(set(it)) in (0, 1)
		and also faster than:
			try:
				first = next(it)
			except StopIteration:
				return True
			return all(first == rest for rest in it)
		for both extreme cases: all equal and none equal
	"""

	g = groupby(it)
	return next(g, True) and not next(g, False) # type: ignore

def pairwise(it):
	# type: (Iterable[T], ) -> Iterator[Tuple[T, T]]

	""" Return two elements of iterable `it` at a time,
		but only advances the iterable by 1 each time.
		(a, b, c) -> (a, b), (b, c)
	"""

	a, b = tee(it, 2)
	next(b, None)
	return zip(a, b)

def findfirst(func, it, default_index=None, default_value=None):
	# type: (Callable, Iterable[T]) -> Optional[T]

	for i, x in enumerate(it):
		if func(x):
			return i, x

	return default_index, default_value

#was: isempty
def is_empty(it):
	# type: (Iterable, ) -> bool

	""" Returns True if the iterable `it` is already fully consumed. """

	try:
		next(it)
		return False
	except StopIteration:
		return True

def consume(it):
	# type: (Iterable, ) -> None

	""" Consumes the iterable `it` completely """

	deque(it, maxlen=0)

def resizer(it, size, pad=False, pad_elm=None):
	# type: (Iterable[Sequence[T]], int, bool, Optional[T]) -> Iterable

	if size <= 0:
		raise ValueError("size must be larger than 0")

	it = iter(it)
	buf = ""
	pos = 0
	try:
		while True:
			to_read = size
			out = []
			while to_read > 0:
				if len(buf) == pos:
					buf = next(it)
					pos = 0
				tmp = buf[pos:pos+to_read]
				out.extend(tmp)
				to_read -= len(tmp)
				pos += len(tmp)
			yield out
	except StopIteration:
		if out:
			while pos < len(buf):
				logger.error("this can never happen, can it?")
				tmp = buf[pos:pos+to_read]
				out.extend(tmp)
				to_read -= len(tmp)
				pos += to_read
			if pad:
				yield out + [pad_elm]*(size-len(out))
			else:
				yield out

def switch(it):
	# type: (Iterable[Tuple[T, U]], ) -> Iterator[Tuple[U, T]]

	""" Swaps the elements in a iterable of pairs.
		((3, 6), (7, 4), (1, 9)) -> (6, 3), (4, 7), (9, 1)
	"""

	return ((j, i) for i, j in it)

def switched_enumerate(it):
	# type: (Iterable[T], ) -> Iterator[Tuple[T, int]]

	return switch(enumerate(it))

def multi_join(it_a, it_b, join_func=add):
	# type: (Iterable[T], Iterable[U], Callable[[T, U], V]) -> Iterator[V]

	""" Calls `join_func` on the cross product of `it_a` and `it_b`.
		multi_join((1, 2), (3, 4), lambda x, y: str(x)+str(y)) -> ('13', '23', '14', '24')
	"""

	return [join_func(a, b) for b in it_b for a in it_a]

def iter_except(iterator, exception_callbacks):
	# type: (Iterator[T], Dict[Type, Callable[[Iterator[T], Exception], bool]]) -> Iterator[T]

	""" Calls callbacks when exceptions are raised in `iterator`. Does not work for Generators.
		The type key in the `exception_callbacks` dict must be the same type as the exception which
		should be caught, not a subtype.
	"""

	if isinstance(iterator, GeneratorType):
		raise TypeError("iterator cannot be a generator")

	while True:
		try:
			yield next(iterator)
		except StopIteration:
			return
		except Exception as e:
			try:
				if exception_callbacks[type(e)](iterator, e): # use if x. and not if not x, so None doesn't raise, if user forgets to returns correct value
					raise
			except KeyError:
				raise e

def iter_except_ignore(iterator, exceptions, logger=logger):
	# type: (Iterator[T], Collection[Exception], Logger) -> Iterator[T]

	""" Ignores `exceptions` raised in `iterator`. Does not work for Generators.
	"""

	if isinstance(iterator, GeneratorType):
		raise TypeError("iterator cannot be a generator")

	while True:
		try:
			yield next(iterator)
		except StopIteration:
			break
		except exceptions as e:
			logger.warning("Error in iterable: %s", str(e))

def decompress(selectors, data, default=None):
	# type: (Iterable[bool], Iterator[T], Optional[T]) -> Iterator[T]

	""" Basically the opposite of `itertools.compress`.
		decompress([True, False, True], iter(["A", "B"])) -> ("A", None, "B")
	"""

	try:
		for s in selectors:
			if s:
				yield next(data)
			else:
				yield default
	except StopIteration: # exception converted because StopIterations raised in generators will cause RuntimeErrors
		raise IteratorExhausted

#was: get_first_non_none
def first_not_none(it):
	# type: (Iterable[T], ) -> Optional[T]

	""" Returns the first element of `it` which is not None. """

	return next(filternone(it), None)

def range_count(start=0, stop=None, step=1):
	# type: (int, Optional[int]) -> Iterator[int]

	""" Similar to `range`, except it optionally allows for an infinite range. """

	if stop:
		return range(start, stop, step)
	else:
		return count(start, step)

def product_range_repeat(depth, args):
	# type: (int, Tuple) -> Iterator

	""" Returns coordinate or coordinate slices useful for iteration over space.
		Can replace higher dimensional for-in-range loops.
	"""

	return product(tuple(range(*args)), repeat=depth)

def powerset(it):
	# type: (Iterable[T], ) -> Iterator[Tuple[T, ...]]

	""" powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3) """

	s = tuple(it)
	return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

def any_in_common(a, b):
	# type: (Iterable, Iterable) -> bool

	""" Tests if iterables `a` and `b` have any elements in common. """

	return any(i == j for i, j in product(a, b))

# had cmp=b"\0" default before
def all_equal_to(iterable, cmp):
	# type: (Iterable[T], T) -> bool

	return all(elm == cmp for elm in iterable)

# was: iter_same
def iter_equal(*its):
	# type: (*Iterable, ) -> bool

	return all(all_equal(it_elements) for it_elements in zip_longest(*its))

def iter_different(it_a, it_b):
	# type: (Iterable, Iterable) -> bool

	""" Test if `it_a` and `it_b` yield different elements. """

	return any(a != b for a, b in zip_longest(it_a, it_b))

def every_n(it, n, pos=0):
	# type: (Iterable[T], int, int) -> T
	""" Yields every `n`-th element from iterable `it` and starts at `pos`. """

	if pos > 0:
		advance(it, pos)

	for i, val in enumerate(it):
		if i % n == 0:
			yield val

def split(it, size):
	# type: (Iterable[T], int) -> Sequence[Iterable[T]]

	""" memory consumption should be proportional to size when output iterators
		are used in parallel loop (zip), else might be higher
	"""

	copies = tee(it, size)
	return tuple(every_n(it, size, pos) for it, pos in zip(copies, range(size)))

def remove_all_dupes(it): # was: remove_dup_list
	# type: (Iterable, ) -> Iterable
	""" Removes all duplicates from `it` while preserving order. """
	# Dave Kirby
	seen = set()
	return (x for x in it if x not in seen and not seen.add(x))

if __name__ == "__main__":
	import timeit
	print(min(timeit.repeat(stmt="ae(range(100000))", setup="from __main__ import all_equal as ae")))

	print(min(timeit.repeat(stmt="ae(itertools.repeat(0, 100000))", setup="import itertools; from __main__ import all_equal as ae", number=1000)))
